"""
Polynomial chaos expansion cross-validation
===========================================
"""

# %%
# Introduction
# ------------
#
# In this example, we show how to perform the cross-validation of the
# :ref:`Ishigami model<use-case-ishigami>` using polynomial chaos expansion.
# More precisely, we use the methods suggested in [muller2016]_, chapter 5, page 257.
# We make the assumption that a dataset is given and we create a metamodel using this data.
# Once done, we want to assess the quality of the metamodel using the data we have.
# Another example of this method is presented in
# :doc:`/auto_numerical_methods/general_methods/plot_pce_design`.
#
# In this example, we compare two methods:
#
# - split the data into two subsets, one for training and one for testing,
# - use k-fold validation.
#
# The split of the data is performed by the `compute_R2_score_by_splitting` function below.
# It uses 75% of the data to estimate the coefficients of the metamodel (this is the training step)
# and use 25% of the data to estimate the :math:`R^2` score (this is the validation step).
# To do this, we use the `split` method of the :class:`~openturns.Sample`.
#
# The K-Fold validation is performed by the `compute_R2_score_by_kfold` function below.
# It uses the K-Fold method with :math:`k = 5`.
# The code uses the :class:`~openturns.KFoldSplitter` class, which computes the appropriate indices.
# Similar results can be obtained with :class:`~openturns.LeaveOneOutSplitter` at a higher cost.
# This cross-validation method is used here to see which polynomial degree
# leads to an accurate metamodel of the Ishigami test function.

# %%
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import ishigami_function

# %%
# Define helper functions
# -----------------------

# %%
# The next function creates a polynomial chaos expansion from
# a training data set and a total degree.

# %%


def compute_sparse_least_squares_chaos(
    inputTrain, outputTrain, multivariateBasis, totalDegree, distribution
):
    """
    Create a sparse polynomial chaos based on least squares.

    * Uses the enumerate rule in multivariateBasis.
    * Uses the LeastSquaresStrategy to compute the coefficients based on
      least squares.
    * Uses LeastSquaresMetaModelSelectionFactory to use the LARS selection method.
    * Uses FixedStrategy in order to keep all the coefficients that the
      LARS method selected.

    Parameters
    ----------
    inputTrain : Sample
        The input design of experiments.
    outputTrain : Sample
        The output design of experiments.
    multivariateBasis : multivariateBasis
        The multivariate chaos multivariateBasis.
    totalDegree : int
        The total degree of the chaos polynomial.
    distribution : Distribution.
        The distribution of the input variable.

    Returns
    -------
    result : PolynomialChaosResult
        The estimated polynomial chaos.
    """
    selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
    projectionStrategy = ot.LeastSquaresStrategy(
        inputTrain, outputTrain, selectionAlgorithm
    )
    enumerateFunction = multivariateBasis.getEnumerateFunction()
    multivariateBasisSize = enumerateFunction.getBasisSizeFromTotalDegree(totalDegree)
    adaptiveStrategy = ot.FixedStrategy(multivariateBasis, multivariateBasisSize)
    chaosAlgo = ot.FunctionalChaosAlgorithm(
        inputTrain, outputTrain, distribution, adaptiveStrategy, projectionStrategy
    )
    chaosAlgo.run()
    result = chaosAlgo.getResult()
    return result


# %%
# The next function computes the :math:`R^2` score by splitting the data set
# into a training set and a test set.


# %%
def compute_R2_score_by_splitting(
    inputSample,
    outputSample,
    multivariateBasis,
    totalDegree,
    distribution,
    split_fraction=0.75,
):
    """
    Compute R2 score by splitting into train/test sets.

    Parameters
    ----------
    inputSample : Sample(size, input_dimension)
        The X dataset.
    outputSample : Sample(size, output_dimension)
        The Y dataset.
    multivariateBasis : multivariateBasis
        The multivariate chaos multivariateBasis.
    totalDegree : int
        The total degree of the chaos polynomial.
    distribution : Distribution.
        The distribution of the input variable.
    split_fraction : float, in (0, 1)
        The proportion of the sample used in the training.

    Returns
    -------
    r2Score : float
        The R2 score.
    """

    training_sample_size = inputSample.getSize()
    inputSampleTrain = ot.Sample(inputSample)  # Make a copy
    outputSampleTrain = ot.Sample(outputSample)
    split_index = int(split_fraction * training_sample_size)
    inputSampleTest = inputSampleTrain.split(split_index)
    outputSampleTest = outputSampleTrain.split(split_index)
    chaosResult = compute_sparse_least_squares_chaos(
        inputSampleTrain,
        outputSampleTrain,
        multivariateBasis,
        totalDegree,
        distribution,
    )
    metamodel = chaosResult.getMetaModel()
    metamodelPredictions = metamodel(inputSampleTest)
    val = ot.MetaModelValidation(outputSampleTest, metamodelPredictions)
    r2Score = val.computeR2Score()
    return r2Score


# %%
# The next function computes the :math:`R^2` score by K-Fold.


def compute_R2_score_by_kfold(
    inputSample,
    outputSample,
    multivariateBasis,
    totalDegree,
    distribution,
    kParameter=5,
):
    """
    Compute R2 score by KFold.

    Parameters
    ----------
    inputSample : Sample(size, input_dimension)
        The X dataset.
    outputSample : Sample(size, output_dimension)
        The Y dataset.
    multivariateBasis : multivariateBasis
        The multivariate chaos multivariateBasis.
    totalDegree : int
        The total degree of the chaos polynomial.
    distribution : Distribution.
        The distribution of the input variable.
    kParameter : int
        The parameter K.

    Returns
    -------
    r2Score : Point(outputDimension)
        The R2 score for each output dimension.
    """
    #
    sampleSize = inputSample.getSize()
    outputDimension = outputSample.getDimension()
    splitter = ot.KFoldSplitter(sampleSize, kParameter)
    r2_score_sample = ot.Sample(0, outputDimension)
    for indicesTrain, indicesTest in splitter:
        inputSampleTrain, inputSampleTest = (
            inputSample[indicesTrain],
            inputSample[indicesTest],
        )
        outputSampleTrain, outputSampleTest = (
            outputSample[indicesTrain],
            outputSample[indicesTest],
        )
        chaosResultKFold = compute_sparse_least_squares_chaos(
            inputSampleTrain,
            outputSampleTrain,
            multivariateBasis,
            totalDegree,
            distribution,
        )
        metamodelKFold = chaosResultKFold.getMetaModel()
        predictionsKFold = metamodelKFold(inputSampleTest)
        val = ot.MetaModelValidation(outputSampleTest, predictionsKFold)
        r2_local = val.computeR2Score()
        r2_score_sample.add(r2_local)
    r2Score = r2_score_sample.computeMean()
    return r2Score


# %%
# Define the training data set
# ----------------------------

# %%
# We start by generating the input and output samples. We use a sample size equal to 200.


# %%
im = ishigami_function.IshigamiModel()
im.distribution.setDescription(["X0", "X1", "X2"])
im.model.setOutputDescription(["$Y$"])
ot.RandomGenerator.SetSeed(0)
sample_size = 200
X = im.distribution.getSample(sample_size)
print("Input sample:")
X[:5]

# %%
Y = im.model(X)
Y.setDescription(["Y"])
print("Output sample:")
Y[:5]


# %%
# We compute a polynomial chaos decomposition with a total degree equal to 5.
# In order to reduce the number of coefficients, the `compute_sparse_least_squares_chaos`
# function creates a sparse chaos using regression and the LARS method.

# %%
dimension = im.distribution.getDimension()
multivariateBasis = ot.OrthogonalProductPolynomialFactory(
    [im.distribution.getMarginal(i) for i in range(dimension)]
)
totalDegree = 5  # Polynomial degree
result = compute_sparse_least_squares_chaos(
    X, Y, multivariateBasis, totalDegree, im.distribution
)
result

# %%
# Get the metamodel.

# %%
metamodel = result.getMetaModel()

# %%
# Validate the metamodel from a test set
# --------------------------------------

# %%
# In order to validate our polynomial chaos, we generate an extra pair of
# input / output samples and use the :class:`~openturns.MetaModelValidation` class.
test_sample_size = 200  # Size of the validation design of experiments
inputTest = im.distribution.getSample(test_sample_size)
outputTest = im.model(inputTest)
metamodelPredictions = metamodel(inputTest)
validation = ot.MetaModelValidation(outputTest, metamodelPredictions)
r2Score = validation.computeR2Score()[0]
graph = validation.drawValidation()
graph.setTitle("R2=%.2f, n=%d" % (r2Score, test_sample_size))
view = otv.View(graph)


# %%
# The plot shows that the score is relatively high and might be satisfactory for some purposes.
# There are however several issues with the previous procedure:
#
# - It may happen that the data in the validation sample with size 200 is more
#   difficult to fit than the data in the training dataset.
#   In this case, the :math:`R^2` score may be pessimistic.
# - It may happen that the data in the validation sample with size 200 is
#   less difficult to fit than the data in the validation dataset.
#   In this case, the :math:`R^2` score may be optimistic.
# - We may not be able to generate an extra dataset for validation.
#   In this case, a part of the original dataset should be used for validation.
# - The polynomial degree may not be appropriate for this data.
# - The dataset may be ordered in some way, so that the split is somewhat arbitrary.
#   One solution to circumvent this is to randomly shuffle the data.
#   This can be done using the :class:`~openturns.KPermutationsDistribution`.
#
# The K-Fold validation aims at solving some of these issues, so that all the
# available data is used in order to estimate the :math:`R^2` score.

# %%
# Compute the R2 score from a test set
# ------------------------------------

# %%
# In the following script, we compute the :math:`R^2` score associated with each polynomial degree from 1 to 10.
split_fraction = 0.75
print(f"Split cross-validation, with {100 * split_fraction:.0f}% for training")
degree_max = 10
degree_list = list(range(1, 1 + degree_max))
n_degrees = len(degree_list)
scoreSampleSplit = ot.Sample(len(degree_list), 1)
for i in range(n_degrees):
    totalDegree = degree_list[i]
    scoreSampleSplit[i] = compute_R2_score_by_splitting(
        X, Y, multivariateBasis, totalDegree, im.distribution, split_fraction
    )
    print(f"Degree = {totalDegree}, score = {scoreSampleSplit[i, 0]:.4f}")


# %%
graph = ot.Graph(
    f"Split CV, {100 * split_fraction:.0f}% for training", "Degree", "$R^2$", True
)
cloud = ot.Cloud(ot.Sample.BuildFromPoint(degree_list), scoreSampleSplit)
cloud.setPointStyle("circle")
graph.add(cloud)
boundingBox = ot.Interval([0.0, 0.0], [1 + degree_max, 1.1])
graph.setBoundingBox(boundingBox)
view = otv.View(graph, figure_kw={"figsize": (5.0, 4.0)})

# %%
# We see that the polynomial degree may be increased up to degree 7,
# after which the :math:`R^2` score does not increase much.

# %%
# Compute the R2 score from K-Fold cross-validation
# -------------------------------------------------
#
# One limitation of the previous method is that the estimate of the
# :math:`R^2` may be sensitive to the particular split of the dataset.
# The following script uses 5-Fold cross validation to estimate the
# :math:`R^2` score.

# %%
kParameter = 5
print(f"{kParameter}-Fold cross-validation")
scoreSampleKFold = ot.Sample(len(degree_list), 1)
for i in range(n_degrees):
    totalDegree = degree_list[i]
    scoreSampleKFold[i] = compute_R2_score_by_kfold(
        X, Y, multivariateBasis, totalDegree, im.distribution, kParameter
    )
    print(f"Degree = {totalDegree}, score = {scoreSampleKFold[i, 0]:.4f}")

# %%
graph = ot.Graph(f"{kParameter}-Fold cross-validation", "Degree", "$R^2$")
cloud = ot.Cloud(ot.Sample.BuildFromPoint(degree_list), scoreSampleKFold)
cloud.setPointStyle("square")
graph.add(cloud)
graph.setBoundingBox(boundingBox)
view = otv.View(graph, figure_kw={"figsize": (5.0, 4.0)})

# %%
# The conclusion is similar to the previous method.

# %%
# Compare the two cross-validation methods.
graph = ot.Graph("CV : split vs K-Fold", "Degree", "$R^2$")
cloud = ot.Cloud(ot.Sample.BuildFromPoint(degree_list), scoreSampleSplit)
cloud.setPointStyle("circle")
cloud.setLegend("Split")
graph.add(cloud)
cloud = ot.Cloud(ot.Sample.BuildFromPoint(degree_list), scoreSampleKFold)
cloud.setPointStyle("square")
cloud.setLegend("K-Fold")
graph.add(cloud)
graph.setLegendPosition("upper left")
graph.setBoundingBox(boundingBox)
view = otv.View(graph, figure_kw={"figsize": (5.0, 4.0)})
# sphinx_gallery_thumbnail_number = 4


# %%
# Conclusion
# ----------
#
# When we select the best polynomial degree which maximizes the :math:`R^2` score,
# the danger is that the validation set is used both for computing the :math:`R^2` and to maximize it:
# hence, the :math:`R^2` score may be optimistic.
# In [muller2016]_, chapter 5, page 269, the authors advocate the split of the dataset into three subsets:
#
# - the training set,
# - the validation set,
# - the test set.
#
# When selecting the best parameters, the validation set is used.
# When estimating the :math:`R^2` score with the best parameters, the test set is used.

# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import math

ot.TESTPREAMBLE()

ot.RandomGenerator.SetSeed(0)

# %%
# Test independentMetropolisHastings on Beta-Binomial conjugate model

# Define Beta-binomial model
a, b, lower, upper = 1.0, 1.0, 0.0, 1.0
n, p = 10, 0.5
prior = ot.Beta(a, b, lower, upper)
model = ot.Binomial(n, p)

# %%
# Simulate data and compute analytical posterior
x = model.getSample(1)
posterior = ot.Beta(a + x[0, 0], b + n - x[0, 0], lower, upper)

# %%
# Define IMH sampler
initialState = [p]
imh_sampler = ot.IndependentMetropolisHastings(
    prior, initialState, ot.Uniform(-1.0, 1.0), [0]
)
slf = ot.SymbolicFunction(["x"], [str(n), "x"])
imh_sampler.setLikelihood(model, x, slf)

# %%
# Generate posterior distribution sample
sampleSize = 10000
xSample = imh_sampler.getSample(sampleSize)

# %%
# Compare empirical to theoretical moments

ott.assert_almost_equal(
    xSample.computeMean(), posterior.getMean(), 0.0, 10.0 / math.sqrt(sampleSize)
)
ott.assert_almost_equal(
    xSample.computeStandardDeviation(),
    posterior.getStandardDeviation(),
    0.0,
    10.0 / math.sqrt(sampleSize),
)

# %%
# Draw the unnormalized probability density
# -----------------------------------------

ot.RandomGenerator.SetSeed(1)
lower_bound = 0.0
print(lower_bound)
upper_bound = 2.0 * math.pi
print(upper_bound)

# %%
# Independent Metropolis-Hastings
# -------------------------------
# Let us use a mixture distribution to approximate the target distribution.
#
# This approximation will serve as the instrumental distribution
# in the independent Metropolis-Hastings algorithm.

exp = ot.Exponential(1.0)
normal = ot.Normal(5.3, 0.4)
instrumentalDistribution = ot.Mixture([exp, normal], [0.9, 0.1])
print(instrumentalDistribution.getWeights())


# %%
# MetropolisHastings classes expect to receive the logarithm to the target density.

log_density = ot.SymbolicFunction(
    "x", "log(2 + sin(x)^2) - (2 + cos(3*x)^3 + sin(2*x)^3) * x"
)

initialState = ot.Point([3.0])  # not important in this case
support = ot.Interval([lower_bound], [upper_bound])
independentMH = ot.IndependentMetropolisHastings(
    log_density, support, initialState, instrumentalDistribution, [0]
)

print(independentMH.getRealization())

# %%
# Get a sample

sampleSize = 1000000
sample = independentMH.getSample(sampleSize)


# %%
# Compute posterior mean

mean_ref = 1.22498
posterior_mean = sample.computeMean()[0]
ott.assert_almost_equal(posterior_mean, mean_ref, 1e-5, 0)

std_ref = 1.61558
posterior_std = sample.computeStandardDeviation()[0]
ott.assert_almost_equal(posterior_std, std_ref, 1e-5, 0)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

coll1 = ot.DistributionCollection(0)
coll1.add(ot.Normal(1.0, 2.5))
coll1.add(ot.Gamma(1.5, 3.0, 0.0))
pointLow = ot.Point(0)
pointLow.add(coll1[0].computeQuantile(0.25)[0])
pointLow.add(coll1[1].computeQuantile(0.25)[0])
pointHigh = ot.Point(0)
pointHigh.add(coll1[0].computeQuantile(0.75)[0])
pointHigh.add(coll1[1].computeQuantile(0.75)[0])
coll2 = ot.DistributionCollection(0)
coll2.add(ot.Gamma(2.5, 2.0, 0.0))
coll2.add(ot.Normal(3.0, 1.5))
# First, check the old constructor
transformation = ot.MarginalTransformationEvaluation(coll1)
print("transformation=", repr(transformation))
print("transformation(", pointLow, ")=", repr(transformation(pointLow)))
print("transformation(", pointHigh, ")=", repr(transformation(pointHigh)))
# Validation using finite difference
eps = 1e-5
factor = 1.0 / (2.0 * eps)
gradientLow = ot.Matrix(5, 2)
gradientHigh = ot.Matrix(5, 2)

# dT/dp0
coll = ot.DistributionCollection(2)
coll[0] = ot.Normal(1.0 + eps, 2.5)
coll[1] = coll1[1]
left = ot.MarginalTransformationEvaluation(coll)
coll[0] = ot.Normal(1.0 - eps, 2.5)
right = ot.MarginalTransformationEvaluation(coll)
dTdp = (left(pointLow) - right(pointLow)) * factor
gradientLow[0, 0] = dTdp[0]
gradientLow[0, 1] = dTdp[1]
dTdp = (left(pointHigh) - right(pointHigh)) * factor
gradientHigh[0, 0] = dTdp[0]
gradientHigh[0, 1] = dTdp[1]
# dT/dp1
coll = ot.DistributionCollection(2)
coll[0] = ot.Normal(1.0, 2.5 + eps)
coll[1] = coll1[1]
left = ot.MarginalTransformationEvaluation(coll)
coll[0] = ot.Normal(1.0, 2.5 - eps)
right = ot.MarginalTransformationEvaluation(coll)
dTdp = (left(pointLow) - right(pointLow)) * factor
gradientLow[1, 0] = dTdp[0]
gradientLow[1, 1] = dTdp[1]
dTdp = (left(pointHigh) - right(pointHigh)) * factor
gradientHigh[1, 0] = dTdp[0]
gradientHigh[1, 1] = dTdp[1]
# dT/dp2
coll = ot.DistributionCollection(2)
coll[0] = coll1[0]
coll[1] = ot.Gamma(1.5 + eps, 3.0, 0.0)
left = ot.MarginalTransformationEvaluation(coll)
coll[1] = ot.Gamma(1.5 - eps, 3.0, 0.0)
right = ot.MarginalTransformationEvaluation(coll)
dTdp = (left(pointLow) - right(pointLow)) * factor
gradientLow[2, 0] = dTdp[0]
gradientLow[2, 1] = dTdp[1]
dTdp = (left(pointHigh) - right(pointHigh)) * factor
gradientHigh[2, 0] = dTdp[0]
gradientHigh[2, 1] = dTdp[1]
# dT/dp3
coll = ot.DistributionCollection(2)
coll[0] = coll1[0]
coll[1] = ot.Gamma(1.5, 3.0 + eps, 0.0)
left = ot.MarginalTransformationEvaluation(coll)
coll[1] = ot.Gamma(1.5, 3.0 - eps, 0.0)
right = ot.MarginalTransformationEvaluation(coll)
dTdp = (left(pointLow) - right(pointLow)) * factor
gradientLow[3, 0] = dTdp[0]
gradientLow[3, 1] = dTdp[1]
dTdp = (left(pointHigh) - right(pointHigh)) * factor
gradientHigh[3, 0] = dTdp[0]
gradientHigh[3, 1] = dTdp[1]
# dT/dp4
coll = ot.DistributionCollection(2)
coll[0] = coll1[0]
coll[1] = ot.Gamma(1.5, 3.0, 0.0 + eps)
left = ot.MarginalTransformationEvaluation(coll)
coll[1] = ot.Gamma(1.5, 3.0, 0.0 - eps)
right = ot.MarginalTransformationEvaluation(coll)
dTdp = (left(pointLow) - right(pointLow)) * factor
gradientLow[4, 0] = dTdp[0]
gradientLow[4, 1] = dTdp[1]
dTdp = (left(pointHigh) - right(pointHigh)) * factor
gradientHigh[4, 0] = dTdp[0]
gradientHigh[4, 1] = dTdp[1]

print(
    "transformation    parameters gradient=",
    repr(transformation.parameterGradient(pointLow)),
)
print("finite difference parameters gradient=", repr(gradientLow))
print(
    "transformation    parameters gradient=",
    repr(transformation.parameterGradient(pointHigh)),
)
print("finite difference parameters gradient=", repr(gradientHigh))
print("input dimension=", transformation.getInputDimension())
print("output dimension=", transformation.getOutputDimension())

# Second, check the constructor for old inverse transformation
transformation = ot.MarginalTransformationEvaluation(
    coll1, ot.MarginalTransformationEvaluation.TO
)
print("transformation=", repr(transformation))
uLow = ot.Point(coll1.getSize(), 0.25)
uHigh = ot.Point(coll1.getSize(), 0.75)
print("transformation(", uLow, ")=", repr(transformation(uLow)))
print("transformation(", uHigh, ")=", repr(transformation(uHigh)))
# Validation using finite difference
# dT/dp0
coll = ot.DistributionCollection(2)
coll[0] = ot.Normal(1.0 + eps, 2.5)
coll[1] = coll1[1]
left = ot.MarginalTransformationEvaluation(coll, ot.MarginalTransformationEvaluation.TO)
coll[0] = ot.Normal(1.0 - eps, 2.5)
right = ot.MarginalTransformationEvaluation(
    coll, ot.MarginalTransformationEvaluation.TO
)
dTdp = (left(uLow) - right(uLow)) * factor
gradientLow[0, 0] = dTdp[0]
gradientLow[0, 1] = dTdp[1]
dTdp = (left(uHigh) - right(uHigh)) * factor
gradientHigh[0, 0] = dTdp[0]
gradientHigh[0, 1] = dTdp[1]
# dT/dp1
coll = ot.DistributionCollection(2)
coll[0] = ot.Normal(1.0, 2.5 + eps)
coll[1] = coll1[1]
left = ot.MarginalTransformationEvaluation(coll, ot.MarginalTransformationEvaluation.TO)
coll[0] = ot.Normal(1.0, 2.5 - eps)
right = ot.MarginalTransformationEvaluation(
    coll, ot.MarginalTransformationEvaluation.TO
)
dTdp = (left(uLow) - right(uLow)) * factor
gradientLow[1, 0] = dTdp[0]
gradientLow[1, 1] = dTdp[1]
dTdp = (left(uHigh) - right(uHigh)) * factor
gradientHigh[1, 0] = dTdp[0]
gradientHigh[1, 1] = dTdp[1]
# dT/dp2
coll = ot.DistributionCollection(2)
coll[0] = coll1[0]
coll[1] = ot.Gamma(1.5 + eps, 3.0, 0.0)
left = ot.MarginalTransformationEvaluation(coll, ot.MarginalTransformationEvaluation.TO)
coll[1] = ot.Gamma(1.5 - eps, 3.0, 0.0)
right = ot.MarginalTransformationEvaluation(
    coll, ot.MarginalTransformationEvaluation.TO
)
dTdp = (left(uLow) - right(uLow)) * factor
gradientLow[2, 0] = dTdp[0]
gradientLow[2, 1] = dTdp[1]
dTdp = (left(uHigh) - right(uHigh)) * factor
gradientHigh[2, 0] = dTdp[0]
gradientHigh[2, 1] = dTdp[1]
# dT/dp3
coll = ot.DistributionCollection(2)
coll[0] = coll1[0]
coll[1] = ot.Gamma(1.5, 3.0 + eps, 0.0)
left = ot.MarginalTransformationEvaluation(coll, ot.MarginalTransformationEvaluation.TO)
coll[1] = ot.Gamma(1.5, 3.0 - eps, 0.0)
right = ot.MarginalTransformationEvaluation(
    coll, ot.MarginalTransformationEvaluation.TO
)
dTdp = (left(uLow) - right(uLow)) * factor
gradientLow[3, 0] = dTdp[0]
gradientLow[3, 1] = dTdp[1]
dTdp = (left(uHigh) - right(uHigh)) * factor
gradientHigh[3, 0] = dTdp[0]
gradientHigh[3, 1] = dTdp[1]
# dT/dp4
coll = ot.DistributionCollection(2)
coll[0] = coll1[0]
coll[1] = ot.Gamma(1.5, 3.0, 0.0 + eps)
left = ot.MarginalTransformationEvaluation(coll, ot.MarginalTransformationEvaluation.TO)
coll[1] = ot.Gamma(1.5, 3.0, 0.0 - eps)
right = ot.MarginalTransformationEvaluation(
    coll, ot.MarginalTransformationEvaluation.TO
)
dTdp = (left(uLow) - right(uLow)) * factor
gradientLow[4, 0] = dTdp[0]
gradientLow[4, 1] = dTdp[1]
dTdp = (left(uHigh) - right(uHigh)) * factor
gradientHigh[4, 0] = dTdp[0]
gradientHigh[4, 1] = dTdp[1]

print(
    "transformation    parameters gradient=",
    repr(transformation.parameterGradient(uLow)),
)
print("finite difference parameters gradient=", repr(gradientLow))
print(
    "transformation    parameters gradient=",
    repr(transformation.parameterGradient(uHigh)),
)
print("finite difference parameters gradient=", repr(gradientHigh))
print("input dimension=", transformation.getInputDimension())
print("output dimension=", transformation.getOutputDimension())

# Third, check the constructor for the new transformation
transformation = ot.MarginalTransformationEvaluation(coll1, coll2)
print("transformation=", repr(transformation))
print("transformation(", pointLow, ")=", repr(transformation(pointLow)))
print("transformation(", pointHigh, ")=", repr(transformation(pointHigh)))
print("input dimension=", transformation.getInputDimension())
print("output dimension=", transformation.getOutputDimension())

T1 = ot.MarginalTransformationEvaluation([ot.Exponential(2.0)], [ot.Exponential(1.0)])
print("T1=", T1)
T2 = ot.MarginalTransformationEvaluation([ot.Exponential(1.0)], [ot.Exponential(2.0)])
print("T2=", T2)
T3 = ot.MarginalTransformationEvaluation(
    [ot.Exponential(3.0, 4.0)], [ot.Exponential(3.0, 5.0)]
)
print("T3=", T3)



import openturns as ot
import openturns.viewer as otv

f = ot.SymbolicFunction(["x"], ["x + x * sin(x)"])
inputSample = ot.Sample([[1.0], [3.0], [5.0], [6.0], [7.0], [8.0]])
outputSample = f(inputSample)

f1 = ot.SymbolicFunction(["x"], ["sin(x)"])
f2 = ot.SymbolicFunction(["x"], ["x"])
f3 = ot.SymbolicFunction(["x"], ["cos(x)"])
basis = ot.Basis([f1, f2, f3])
covarianceModel = ot.SquaredExponential([1.0])
covarianceModel.setActiveParameter([])
algo = ot.GeneralLinearModelAlgorithm(inputSample, outputSample, covarianceModel, basis)
algo.run()
result = algo.getResult()
metamodel = result.getMetaModel()

graph = f.draw(0.0, 10.0)
graph.add(metamodel.draw(0.0, 10.0))
graph.add(ot.Cloud(inputSample, outputSample))
graph.setLegends(["model", "meta model", "sample"])
graph.setLegendPosition("upper left")
graph.setTitle(r"$x + x \; \sin(x)$")
otv.View(graph)



"""
Fit a distribution by maximum likelihood
========================================
"""

# %%
# In this example we are going to estimate the parameters of a parametric by generic numerical optimization of the likelihood.
#
# The likelihood of a sample :math:`(\vect{x}_1, \dots, \vect{x}_n)` according to a parametric density function :math:`p_{\vect{\theta}}` is:
#
# .. math::
#    \ell(\vect{x}_1, \dots, \vect{x}_n,\vect{\theta}) = \prod_{i=1}^n p_{\vect{\theta}}(\vect{x}_i)
#

# %%
import openturns as ot
import math as m


# %%
# Create data from the PDF of the Normal distribution with :math:`\mu=4`, :math:`\sigma=1.5`.
sample = ot.Normal(4.0, 1.5).getSample(200)

# %%
# Create the search interval of (:math:`\mu`, :math:`\sigma`) : the constraint is :math:`\sigma>0`.
lowerBound = [-1.0, 1.0e-4]
upperBound = [-1.0, -1.0]
finiteLowerBound = [False, True]
finiteUpperBound = [False, False]
bounds = ot.Interval(lowerBound, upperBound, finiteLowerBound, finiteUpperBound)

# %%
# Create the starting point of the research:
#
# - for :math:`\mu` : the first point,
# - for :math:`\sigma` : a value evaluated from the two first points.
mu0 = sample[0][0]
sigma0 = m.sqrt((sample[1][0] - sample[0][0]) * (sample[1][0] - sample[0][0]))
startingPoint = [mu0, sigma0]

# %%
# Create the estimator from a parametric PDF.
pdf = ot.Normal()
factory = ot.MaximumLikelihoodFactory(pdf)
factory.setOptimizationBounds(bounds)

# %%
# Set the starting point via the solver.
solver = factory.getOptimizationAlgorithm()
solver.setStartingPoint(startingPoint)
factory.setOptimizationAlgorithm(solver)

# %%
# Estimate the parametric model.
distribution = factory.build(sample)
str(distribution)

# %%
# Retrieve the estimated parameters.
print(distribution.getParameter())



"""
Defining Python and symbolic functions: a quick start introduction to functions
===============================================================================
"""

# %%
# Abstract
# --------
#
# In this example, we show how to define Python and symbolic functions.
# Such functions can be evaluated by the library and used, for example, to propagate uncertainties.
# We focus on functions which have a vector input and a vector output.

# %%
# Introduction
# ------------
#
# We consider the following example:
#
# * three input variables,
# * two outputs.
#
# Moreover, we assume that the input distribution has independent Gaussian marginals.
#
# The function is defined by the equations:
#
# .. math::
#    Y_1 = X_1 + X_2 + X_3
#
# and
#
# .. math::
#    Y_2 = X_1 - X_2 X_3
#
#
# for any :math:`X_1,X_2,X_3 \in \mathbb{R}`.
#
# The exact expectation and standard deviation of the output random variable are presented in the following table.
#
# =============     ===========  ==================
# Variable          Expectation  Standard deviation
# =============     ===========  ==================
#  :math:`Y_1`       0            1.732
#  :math:`Y_2`       0            1.415
# =============     ===========  ==================
#

# %%
import numpy as np
import openturns as ot


# %%
# We first define the input random vector of the function.

# %%
X0 = ot.Normal(0.0, 1.0)
X1 = ot.Normal(0.0, 1.0)
X2 = ot.Normal(0.0, 1.0)
inputDistribution = ot.JointDistribution((X0, X1, X2))
inputRandomVector = ot.RandomVector(inputDistribution)


# %%
# The Python function
# -------------------
#
# Based on a Python function defined with the `def` keyword, the `PythonFunction` class creates a function.
#
# The simplest constructor of the `PythonFunction` class is:
#
# `PythonFunction ( nbInputs , nbOutputs , myPythonFunc )`
#
# where
#
# * `nbInputs`: the number of inputs,
# * `nbOutputs`: the number of outputs,
# * `myPythonFunc`: a Python function.
#
# The goal of the `PythonFunction` class are:
#
# * to easily create a function in Python,
# * use all the power of the Python libraries in order to evaluate the function.

# %%
# The function `mySimulator` has the calling sequence `y=mySimulator(x)` where:
#
# * `x`: the input of the function, a vector with `nbInputs` dimensions,
# * `y`: the output of the function, a vector with `nbOutputs` dimensions.


# %%
def mySimulator(x):
    y0 = x[0] + x[1] + x[2]
    y1 = x[0] - x[1] * x[2]
    y = [y0, y1]
    return y


# %%
# We now define the `PythonFunction` object.

# %%
myfunction = ot.PythonFunction(3, 2, mySimulator)

# %%
# This function can be evaluated using parentheses. It produces the same outputs as the `mySimulator` function.

# %%
myfunction([1.0, 2.0, 3.0])

# %%
# However, the newly created `myfunction` has services that the basic Python function did not have.
# For example, we can create a `CompositeRandomVector` on top of it, by associating it to the input random vector.

# %%
outputVect = ot.CompositeRandomVector(myfunction, inputRandomVector)

# %%
# In the following example, we estimate the output mean based on a simple Monte-Carlo experiment using 10000 function evaluations.

# %%
montecarlosize = 10000
outputSample = outputVect.getSample(montecarlosize)

# %%
empiricalMean = outputSample.computeMean()
print(empiricalMean)
empiricalSd = outputSample.computeStandardDeviation()
print(empiricalSd)

# %%
# What types for x and for y ?
# ----------------------------
#
# Not all types are possible for the inputs and outputs of the `mySimulator` function.
# The following table present some of the available types.
# All in all, any type which can be converted to or from a "vector" can be managed by the `PythonFunction` class.
#
#  ====================  =======  ========
#  Type                  Input X  Output Y
#  ====================  =======  ========
#  `list` (Python)       NO       YES
#  `tuple` (Python)      NO       YES
#  `array` (NumPy)       NO       YES
#  `Point` (OpenTURNS)   YES      YES
#  ====================  =======  ========
#

# %%
# The vectorized Python function
# ------------------------------
#
# The `PythonFunction` class has a `func_sample` option which vectorizes the
# computation so that all the output values in the sample are computed from a
# single function call, without any `for` loop.
# To make this possible, the input and output is then a `Sample` instead of a `Point`.
# This often boosts performance (but not always).

# %%
# The calling sequence of a vectorized Python function is:
#
# def mySimulator (x):
#     [...]
#     return y
#
# myfunction = PythonFunction(nbInputs, nbOutputs, func_sample = mySimulator)
#
# where
#
# * x: the input of the function, a `Sample` with size `nbExperiments` (`getSize`) and dimension `nbInputs` (`getDimension`),
# * y: the output of the function
#
#   * a numpy `array` with `nbExperiments` rows and `nbOutputs` columns
#   * or a `Sample` with size `nbExperiments` and dimension `nbOutputs`

# %%
# In the following, we present a vectorization example based on the `numpy` module.

# %%


# %%
def mySimulatorVect(x):
    # Convert NumericalSample > Array Numpy
    x = np.array(x)
    x0 = x[:, 0]  # Extract column 0
    x1 = x[:, 1]
    x2 = x[:, 2]
    y0 = x0 + x1 + x2
    y1 = x0 - x1 * x2
    # Stack the two rows
    y = np.vstack((y0, y1))
    y = y.transpose()
    return y


# %%
myfunctionVect = ot.PythonFunction(3, 2, func_sample=mySimulatorVect)

# %%
outputVect = ot.CompositeRandomVector(myfunctionVect, inputRandomVector)

# %%
montecarlosize = 10000
outputSample = outputVect.getSample(montecarlosize)

empiricalMean = outputSample.computeMean()
print(empiricalMean)
empiricalSd = outputSample.computeStandardDeviation()
print(empiricalSd)

# %%
# How to manage the history
# -------------------------
#
# The :class:`openturns.MemoizeFunction` class defines a history system to store the calls to the function.
#
#  ====================   ===============================================
#  Methods                Description
#  ====================   ===============================================
#  `enableHistory()`      enables the history (it is enabled by default)
#  `disableHistory()`     disables the history
#  `clearHistory()`       deletes the content of the history
#  `getInputHistory()`    a `Sample`, the history of inputs X
#  `getOutputHistory()`   a `Sample`, the history of outputs Y
#  ====================   ===============================================

# %%
myfunction = ot.PythonFunction(3, 2, mySimulator)
myfunction = ot.MemoizeFunction(myfunction)

# %%
outputVariableOfInterest = ot.CompositeRandomVector(myfunction, inputRandomVector)
montecarlosize = 10
outputSample = outputVariableOfInterest.getSample(montecarlosize)

# %%
# Get the history of input points.

# %%
inputs = myfunction.getInputHistory()
inputs

# %%
# Symbolic functions
# ------------------
#
# The `SymbolicFunction` class can create symbolic functions whenever the function is simple enough to be expressed as a string.
#
# This has at least two significant advantages.
#
# * It generally improves the performance.
# * This allows one to automatically evaluate the exact gradient and Hessian matrix.
#
# In practice, some functions cannot be expressed as a symbolic function: in this case, the Python function cannot be avoided.

# %%
# The calling sequence is the following:
#
# `
# myfunction = SymbolicFunction( list_of_inputs, list_of_formulas)
# `
#
# where
#
# * list_of_inputs: a `list` of `nbInputs` strings, the names of the input variables,
# * list_of_formulas: a `list` of `nbOutputs` strings, the equations.

# %%
myfunction = ot.SymbolicFunction(("x0", "x1", "x2"), ("x0 + x1 + x2", "x0 - x1 * x2"))

# %%
# A `SymbolicFunction`, like any other function, can also have a history.

# %%
myfunction = ot.MemoizeFunction(myfunction)

# %%
outputVect = ot.CompositeRandomVector(myfunction, inputRandomVector)

# %%
montecarlosize = 10000
outputSample = outputVect.getSample(montecarlosize)
empiricalMean = outputSample.computeMean()
print(empiricalMean)

# %%
# Since the history is enabled, we can get the history of outputs of the function.

# %%
outputs = myfunction.getOutputHistory()
outputs[1:10, :]



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


matrix1 = ot.SymmetricMatrix(2)
matrix1.setName("matrix1")
matrix1[0, 0] = 1.0
matrix1[1, 0] = 5.0
matrix1[1, 1] = 12.0
print("matrix1=", matrix1)
print("matrix1=")
print(matrix1.__str__())

pt = ot.Point(0)
pt.add(5.0)
pt.add(0.0)
print("pt=", pt)

result = matrix1.solveLinearSystem(pt)
print("result=", result)
#    print "verif. ", matrix1 * result - pt

determinant = matrix1.computeDeterminant()
print("determinant= %.1f" % determinant)

b = ot.Matrix(2, 3)
b[0, 0] = 5.0
b[1, 0] = 0.0
b[0, 1] = 10.0
b[1, 1] = 1.0
b[0, 2] = 15.0
b[1, 2] = 2.0
result2 = ot.Matrix()
result2 = matrix1.solveLinearSystem(b)
print("result2=", result2)
print("result2=")
print(result2.__str__())

ev = matrix1.computeEigenValues()
print("ev=", ev)

ev, evect = matrix1.computeEV()
print("ev=", ev)
print("evect=", repr(evect))
print("evect=")
print(evect.__str__())
maxModule = matrix1.computeLargestEigenValueModule(10, 1e-2)
print("max |ev|=%.6g" % maxModule)

# Check the high dimension determinant computation
matrix3 = ot.SymmetricMatrix(3)
matrix3[0, 0] = 1.0
matrix3[0, 1] = 2.0
matrix3[0, 2] = 3.0
matrix3[1, 1] = 2.5
matrix3[1, 2] = -3.5
matrix3[2, 2] = 2.5

print("matrix3=")
print(matrix3.__str__())
# sign = 0.0
# value = matrix3.computeLogAbsoluteDeterminant(sign)
# print "log(|det|)=", value, ", sign=", sign
value = matrix3.computeDeterminant()
print("det=", value)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Default constructor */
point1 = ot.Point()
ref_point1 = point1

# Check method add()
point1.add(0.0)
point1.add(1.0)

size = ref_point1.getDimension()
print("size of point1 = ", size)

val1 = ref_point1[0]
val2 = ref_point1[1]
print("point1 = ", repr(ref_point1))
print("point1[0] = %.6f" % val1)
print("point1[1] = %.6f" % val2)

# Constructor with size
point2 = ot.Point(2)
ref_point2 = point2

# Check operator[] methods
point2[0] = 10.0
point2[1] = 11.0

val1 = ref_point2[0]
val2 = ref_point2[1]
print("point2[0] = %.6f" % val1)
print("point2[1] = %.6f" % val2)

# Copy constructor
point3 = ot.Point(ref_point1)
ref_point3 = point3

val1 = ref_point3[0]
val2 = ref_point3[1]
print("point3[0] = %.6f" % val1)
print("point3[1] = %.6f" % val2)

# Constructor from indices
indices = ot.Indices(5)
indices.fill()
point = ot.Point(indices)
print("point=", point)

# Constructor from python sequence
point4 = ot.Point((1000.0, 1001.0))
i = 0
for coord in point4:
    print("point4[%d] = %.6f" % (i, coord))
    i += 1

# Comparison operator
if ref_point2 != ref_point2:
    print(
        "OT::Base::Type::Point.operator == does NOT return the correct value. Says that point2 and point2 are DIFFERENT though they are equal."
    )
    raise RuntimeError()

if ref_point2 == ref_point3:
    print(
        "OT::Base::Type::Point.operator == does NOT return the correct value. Says that point2 and point3 are EQUAL though they are different."
    )
    raise RuntimeError()

# Addition/Subtraction/Product operators
point5 = ref_point1 + ref_point2
ref_point5 = point5
print("point5 = ", repr(ref_point5))

point6 = ref_point1 - ref_point2
ref_point6 = point6
print("point6 = ", repr(ref_point6))

point7 = ref_point5
point7 += ref_point1 + ref_point2
ref_point7 = point7
print("point7 = ", repr(ref_point7))

point8 = ref_point6
point8 -= ref_point1 - ref_point2
ref_point8 = point8
print("point8 = ", repr(ref_point8))

point9 = ot.Point(2, 5.0) / 2.0
print("point9 = ", point9)

point10 = ot.Point(2, 5.0)
point10 /= 2.0
print("point10 = ", point10)

# We create an orthonormal base (O,i,j)
i = ot.Point(2)
i[0] = 1
i[1] = 0
print("i = ", repr(i))

j = ot.Point(2)
j[0] = 0
j[1] = 1
print("j = ", repr(j))

# BUG PYTHON for I = Point(2. * i)
II = ot.Point(i * 2.0)
print("I = ", repr(II))

J = ot.Point(j * 3.0)
print("J = ", repr(J))

dotProduct = II.dot(J)
print("dotProduct = %.6f" % dotProduct)

# slicing
point8 = ot.Point((1.0, 2.0, 3.0, 4.0, 5.0))
print("point8[2] = ", point8[2])
print("point8[1:3] = ", point8[1:3])
print("point8[:] = ", point8[:])
point8[1:3] = (88, 99)
print("point8 = ", point8)
point8[0:5] = ot.Point(5, 9)
print("point8 = ", point8)

# equality
pt1 = ot.Point([1.1, 2.2, 3.3])
pt2 = ot.Point((1.1, 2.2, 3.3))
print("Equality ? ", (pt1 == pt2))

# non-regression for #754
pt = ot.Point([-3, -5, 6])
print("pt = ", pt)

# last
print("pt[-1]=", pt[-1])
pt[-1] = 8.0
print("pt[-1]=", pt[-1])

print(0 + pt1)
try:
    1 + pt1
except Exception:
    print("1+Point -> ok")

# unary minus
x = ot.Point([1, 2, 3])
print(-x)



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


# Create a bivariate Gaussian process
myMesh = ot.IntervalMesher([39, 39]).build(ot.Interval([0.0] * 2, [1.0] * 2))
myCov = ot.GeneralizedExponential(2 * [0.1], 1.3)
myProcess = ot.GaussianProcess(myCov, myMesh)

myField = myProcess.getRealization()

graph = myField.drawMarginal(0, False)
graph.setTitle("A field")

fig = plt.figure(figsize=(8, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=True)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

dimension = 3

# Create the orthogonal basis
enumerateFunction = ot.LinearEnumerateFunction(dimension)
productBasis = ot.OrthogonalProductPolynomialFactory(
    [ot.LegendreFactory()] * dimension, enumerateFunction
)

degree = 6
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(degree)
maximumDimension = 100
maximumSize = 20
significanceFactor = 1e-4
adaptiveStrategy = ot.CleaningStrategy(
    productBasis, maximumDimension, maximumSize, significanceFactor
)

assert adaptiveStrategy.involvesModelSelection() is True

adaptiveStrategy.computeInitialBasis()
assert adaptiveStrategy.getCurrentVectorIndex() == 20

psi = adaptiveStrategy.getPsi()
assert len(psi) == 20

alpha_k = [3.5, 0.1, 0.0, -0.2, 0.0, 0.3, 0.0, -0.4, 0.0, -0.5]
residual = 0.0
relativeError = 0.0
adaptiveStrategy.updateBasis(alpha_k, residual, relativeError)

psi = adaptiveStrategy.getPsi()
assert len(psi) == 7
assert adaptiveStrategy.getCurrentVectorIndex() == 21



#!/usr/bin/env python

import openturns as ot
from openturns.usecases import ishigami_function
import openturns.viewer as otv

# data
im = ishigami_function.IshigamiModel()
N = 1000
g = im.model
x = im.distribution.getSample(N)
P = x.getDimension()
marginals = [im.distribution.getMarginal(i) for i in range(P)]
y = g(x)

# polynomial chaos
q, totalDegree = 0.4, 5
enumerateFunction = ot.HyperbolicAnisotropicEnumerateFunction(P, q)
productBasis = ot.OrthogonalProductPolynomialFactory(marginals, enumerateFunction)
approximationAlgorithm = ot.LeastSquaresMetaModelSelectionFactory(
    ot.LARS(), ot.CorrectedLeaveOneOut()
)
adaptiveStrategy = ot.FixedStrategy(
    productBasis, enumerateFunction.getStrataCumulatedCardinal(totalDegree)
)
projectionStrategy = ot.LeastSquaresStrategy(approximationAlgorithm)
algo = ot.FunctionalChaosAlgorithm(
    x, y, im.distribution, adaptiveStrategy, projectionStrategy
)
algo.run()
result = algo.getResult()
graph = result.drawSelectionHistory()
otv.View(graph)



"""
Optimization using NLopt
========================
"""

# %%
# In this example we are going to explore optimization using the interface to the `NLopt <https://nlopt.readthedocs.io/en/latest/>`_ library.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# List available algorithms
for algo in ot.NLopt.GetAlgorithmNames():
    print(algo)

# %%
# More details on NLopt algorithms are available `here <https://nlopt.readthedocs.io/en/latest/NLopt_Algorithms/>`_ .

# %%
# The optimization algorithm is instantiated from the NLopt name
algo = ot.NLopt("LD_SLSQP")

# %%
# define the problem
objective = ot.SymbolicFunction(["x1", "x2"], ["100*(x2-x1^2)^2+(1-x1)^2"])
inequality_constraint = ot.SymbolicFunction(["x1", "x2"], ["x1-2*x2"])
dim = objective.getInputDimension()
bounds = ot.Interval([-3.0] * dim, [5.0] * dim)
problem = ot.OptimizationProblem(objective)
problem.setMinimization(True)
problem.setInequalityConstraint(inequality_constraint)
problem.setBounds(bounds)

# %%
# solve the problem
algo.setProblem(problem)
startingPoint = [0.0] * dim
algo.setStartingPoint(startingPoint)
algo.run()

# %%
# retrieve results
result = algo.getResult()
print("x^=", result.getOptimalPoint())

# %%
# draw optimal value history
graph = result.drawOptimalValueHistory()
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



import openturns as ot
import openturns.viewer as otv


f = ot.SymbolicFunction(["x"], ["sin(x)"])
a = -2.5
b = 4.5
# Graph
g = f.draw(a, b, 512)

# Fejer type 1
algo = ot.FejerAlgorithm([20], ot.FejerAlgorithm.FEJERTYPE1)
value, nodes = algo.integrateWithNodes(f, ot.Interval(a, b))
lower = ot.Cloud(nodes, ot.Sample(nodes.getSize(), 1))
lower.setColor("green")
lower.setPointStyle("circle")
g.add(lower)

# Fejer type 2
algo = ot.FejerAlgorithm([20], ot.FejerAlgorithm.FEJERTYPE2)
value, nodes = algo.integrateWithNodes(f, ot.Interval(a, b))
lower = ot.Cloud(nodes, ot.Sample(nodes.getSize(), [-1.0 / 8]))
lower.setColor("red")
lower.setPointStyle("square")
g.add(lower)

# Clenshaw-Curtis
algo = ot.FejerAlgorithm([20], ot.FejerAlgorithm.CLENSHAWCURTIS)
value, nodes = algo.integrateWithNodes(f, ot.Interval(a, b))
lower = ot.Cloud(nodes, ot.Sample(nodes.getSize(), [1.0 / 8]))
lower.setColor("magenta")
lower.setPointStyle("plus")
g.add(lower)

g.setTitle(r"Fejer algorithms example: $\int_{-5/2}^{9/2}\sin(t)\,dt=$" + str(value[0]))

g.setLegends(["f", "Fejer_1", "Fejer_2", "Clenshaw-Curtis"])
g.setLegendPosition("upper right")

otv.View(g, figure_kw={"figsize": (8, 4)}, add_legend=True)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


jacobi = ot.JacobiFactory(2.5, 3.5, ot.JacobiFactory.PROBABILITY)
print("jacobi=", jacobi)
jacobi = ot.JacobiFactory(2.5, 3.5, ot.JacobiFactory.ANALYSIS)
print("jacobi=", jacobi)
jacobi = ot.JacobiFactory(2.5, 3.5)
print("jacobi=", jacobi)
for i in range(10):
    print("jacobi(", i, ")=", jacobi.build(i))
roots = jacobi.getRoots(10)
print("jacobi(10) roots=", repr(roots))
nodes, weights = jacobi.getNodesAndWeights(10)
print("jacobi(10) nodes=", nodes, "and weights=", weights)



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

pdf_graph = ot.Graph("PDF graph", "x", "PDF")
pdf_graph.setLegendPosition("upper left")
cdf_graph = ot.Graph("CDF graph", "x", "CDF")
cdf_graph.setLegendPosition("upper left")
palette = ot.Drawable.BuildDefaultPalette(10)
for i, p in enumerate([(1, 2), (2, 1), (5, 2), (10, 1), (100, 100)]):
    d1, d2 = p
    distribution = ot.FisherSnedecor(d1, d2)
    pdf_curve = distribution.drawPDF().getDrawable(0)
    cdf_curve = distribution.drawCDF().getDrawable(0)
    pdf_curve.setColor(palette[i])
    cdf_curve.setColor(palette[i])
    legend = f"d1={d1}, d2={d2}"
    pdf_curve.setLegend(legend)
    cdf_curve.setLegend(legend)
    pdf_graph.add(pdf_curve)
    cdf_graph.add(cdf_curve)
fig = plt.figure(figsize=(10, 4))
pdf_axis = fig.add_subplot(121)
cdf_axis = fig.add_subplot(122)
otv.View(pdf_graph, figure=fig, axes=[pdf_axis], add_legend=True)
otv.View(cdf_graph, figure=fig, axes=[cdf_axis], add_legend=True)
fig.suptitle("FisherSnedecor(d1, d2)")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


distribution = ot.Uniform(1.0, 2.5)
size = 10000
sample = distribution.getSample(size)
factory = ot.UniformFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedUniform = factory.buildAsUniform(sample)
print("Uniform          =", distribution)
print("Estimated uniform=", estimatedUniform)
estimatedUniform = factory.buildAsUniform()
print("Default uniform=", estimatedUniform)
estimatedUniform = factory.buildAsUniform(distribution.getParameter())
print("Uniform from parameters=", estimatedUniform)



"""
Create a maximum entropy order statistics distribution
======================================================
"""

# %%
# Context
# -------
#
# In this example we illustrate how to build the joint distribution of the
# random vector :math:`\inputRV = (X_1, \dots, X_n)` which marginals are given
# and such that:
#
# .. math::
#    :label: orderConstraintEx
#
#    X_1 \leq \dots \leq X_n \quad \mbox{almost surely}.
#
# The class :class:`~openturns.MaximumEntropyOrderStatisticsDistribution` implements the joint
# distribution of :math:`\inputRV` that maximizes the entropy of  :math:`\inputRV`.
# See [fischer2017]_ for more details.
#
# The distribution of :math:`\inputRV` is the distribution of the :math:`n` order statistics of a
# distribution :math:`Y`. We can write:
#
# .. math::
#   (X_1, \dots, X_n) \sim (Y_{(1)}, \dots, Y_{(n)})
#
# In that case, we do not know  :math:`Y` but we fix the marginals of its order
# statistics distribution. If the :math:`X_i \sim \beta(i, n-i+1)`, then :math:`\inputRV` is the
# joint distribution of the :math:`n` order statistics of :math:`Y \sim \cU([0,1])`.
#
# The marginals of  :math:`\inputRV` must be compatible with the order constraint :eq:`orderConstraint`.
# The library checks this compatibility by default. It requires that if :math:`F_i`
# is the CDF of :math:`X_i`, then they must verify:
#
# .. math::
#    :label: orderConstraintCDFEx
#
#    F_1 \geq F_2 \geq \dots \geq F_n
#
# which implies that if :math:`[a_i, b_i]` is the numerical range of :math:`X_i`, then:
#
# .. math::
#    a_i \leq a_{i+1}\\
#    b_i \leq b_{i+1}

# %%
import openturns as ot
import openturns.viewer as otv

ot.ResourceMap.SetAsString("Contour-DefaultColorMapNorm", "rank")


# %%
# Create the distributions :math:`X_i`:
coll = [ot.Beta(1.5, 1.7, 0.0, 1.0), ot.Beta(2.0, 2.3, 0.5, 1.2)]

# %%
# Create the joint order statistics distribution with maximum entropy: the creation is a
# success, which means that the marginals are compatible with the order constraint
# :eq:`orderConstraintEx` and fulfill :eq:`orderConstraintCDFEx`.
jointDist = ot.MaximumEntropyOrderStatisticsDistribution(coll)

# %%
# Draw a sample: the components are well ordered.
x_sample = jointDist.getSample(10)
print(x_sample)

# %%
# Draw PDF
graph = jointDist.drawPDF()
contour = graph.getDrawable(0).getImplementation()
contour.setIsFilled(True)
contour.buildDefaultLevels(50)
graph.setDrawable(0, contour)
graph.setTitle("PDF of the Maximum Entropy Order Statistics Distribution")
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import ishigami_function

ot.TESTPREAMBLE()


# Problem parameters
dimension = 3
im = ishigami_function.IshigamiModel()

size = 1000
x = im.distribution.getSample(size)
y = im.model(x)

# To reduce the time needed by the test
ot.ResourceMap.SetAsUnsignedInteger("FittingTest-LillieforsMinimumSamplingSize", 4)
ot.ResourceMap.SetAsUnsignedInteger("FittingTest-LillieforsMaximumSamplingSize", 4)
ot.ResourceMap.SetAsBool("FunctionalChaosAlgorithm-Sparse", True)
algo = ot.FunctionalChaosAlgorithm(x, y)
algo.run()
result = algo.getResult()
sensitivity = ot.FunctionalChaosSobolIndices(result)

print(sensitivity)
print(sensitivity._repr_html_())
print(sensitivity.__repr_markdown__())

# Create consistent PCE of Ishigami
# build basis
degree = 10
enumerateFunction = ot.LinearEnumerateFunction(dimension)
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(degree)
productBasis = ot.OrthogonalProductPolynomialFactory(
    [ot.LegendreFactory()] * dimension, enumerateFunction
)
adaptiveStrategy = ot.FixedStrategy(productBasis, basisSize)
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
projectionStrategy = ot.LeastSquaresStrategy(selectionAlgorithm)
x = im.distribution.getSample(size)
y = im.model(x)
algo = ot.FunctionalChaosAlgorithm(
    x, y, im.distribution, adaptiveStrategy, projectionStrategy
)
algo.run()
result = algo.getResult()
sensitivity = ot.FunctionalChaosSobolIndices(result)


#
def printSobolResult(S_computed, S_exact):
    absolute_error = abs(S_computed - S_exact)
    print(
        "   S (PCE)    = %.4f" % (S_computed),
        "   S (exact)  = %.4f" % (S_exact),
        "   Abs. Error = %.4e" % (absolute_error),
    )
    return None


#
rtol = 0.0
atol = 0.001
#
print("1. Check first order Sobol' indices")
print("First order, X1")
S_computed = sensitivity.getSobolIndex(0)
printSobolResult(S_computed, im.S1)
assert_almost_equal(S_computed, im.S1, rtol, atol)
#
print("First order, X2")
S_computed = sensitivity.getSobolIndex(1)
printSobolResult(S_computed, im.S2)
assert_almost_equal(S_computed, im.S2, rtol, atol)
#
print("First order, X3")
S_computed = sensitivity.getSobolIndex(2)
printSobolResult(S_computed, im.S3)
assert_almost_equal(S_computed, im.S3, rtol, atol)
#
#
print("2. Check total Sobol' indices")
print("Total, X1")
S_computed = sensitivity.getSobolTotalIndex(0)
printSobolResult(S_computed, im.ST1)
assert_almost_equal(S_computed, im.ST1, rtol, atol)
#
print("Total, X2")
S_computed = sensitivity.getSobolTotalIndex(1)
printSobolResult(S_computed, im.ST2)
assert_almost_equal(S_computed, im.ST2, rtol, atol)
#
print("Total, X3")
S_computed = sensitivity.getSobolTotalIndex(2)
printSobolResult(S_computed, im.ST3)
assert_almost_equal(S_computed, im.ST3, rtol, atol)
#
#
print("3. Check first order (closed) Sobol' indices of groups")
group = [0]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
printSobolResult(S_computed, im.S1)
assert_almost_equal(S_computed, im.S1, rtol, atol)
#
group = [1]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
printSobolResult(S_computed, im.S2)
assert_almost_equal(S_computed, im.S2, rtol, atol)
#
group = [2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
printSobolResult(S_computed, im.S3)
assert_almost_equal(S_computed, im.S3, rtol, atol)
#
group = [0, 1]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
S_exact = im.S1 + im.S2 + im.S12
printSobolResult(S_computed, S_exact)
assert_almost_equal(S_computed, S_exact, rtol, atol)
#
group = [1, 2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
S_exact = im.S2 + im.S3 + im.S23
printSobolResult(S_computed, S_exact)
assert_almost_equal(S_computed, S_exact, rtol, atol)
#
group = [0, 2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
S_exact = im.S1 + im.S3 + im.S13
printSobolResult(S_computed, S_exact)
assert_almost_equal(S_computed, S_exact, rtol, atol)
#
group = [0, 1, 2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedIndex(group)
printSobolResult(S_computed, 1.0)
assert_almost_equal(S_computed, 1.0, rtol, atol)

#
print("4. Check total order index of a group")
group = [0]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
printSobolResult(S_computed, im.ST1)
assert_almost_equal(S_computed, im.ST1, rtol, atol)
#
group = [1]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
printSobolResult(S_computed, im.ST2)
assert_almost_equal(S_computed, im.ST2, rtol, atol)
#
group = [2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
printSobolResult(S_computed, im.ST3)
assert_almost_equal(S_computed, im.ST3, rtol, atol)
#
group = [0, 1]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
S_exact = im.S1 + im.S2 + im.S12 + im.S13 + im.S123
printSobolResult(S_computed, S_exact)
assert_almost_equal(S_computed, S_exact, rtol, atol)
#
group = [1, 2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
S_exact = im.S2 + im.S3 + im.S23 + im.S13 + im.S123
printSobolResult(S_computed, S_exact)
assert_almost_equal(S_computed, S_exact, rtol, atol)
#
group = [0, 2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
S_exact = im.S1 + im.S3 + im.S13 + im.S23 + im.S123
printSobolResult(S_computed, S_exact)
assert_almost_equal(S_computed, S_exact, rtol, atol)
#
group = [0, 1, 2]
print("Group = ", group)
S_computed = sensitivity.getSobolGroupedTotalIndex(group)
printSobolResult(S_computed, 1.0)
assert_almost_equal(S_computed, 1.0, rtol, atol)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

g = ot.SymbolicFunction(["t", "x"], ["x + t^2", "5*x + 3*t"])
grid = ot.RegularGrid(0.0, 0.2, 6)
f = ot.VertexValuePointToFieldFunction(g, grid)

print("f=", f)
# Get the input and output description
print("f input description=", f.getInputDescription())
print("f output description=", f.getOutputDescription())
# Get the input and output dimension, based on description
print("f input dimension=", f.getInputDimension())
print("f output dimension=", f.getOutputDimension())
x = [4.0]
print("input=", x)
print("output time series=", f(x))
# Get the number of calls
print("called ", f.getCallsNumber(), " times")

x = [[3.0], [4.0], [5.0]]
print("input=", x)
print("output time series=", f(x))
# Get the number of calls
print("called ", f.getCallsNumber(), " times")
f1 = f.getMarginal(1)
print("f1=", f1)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# First constructor: distribution of the maximum of the components
# of a random vector given its joint distribution
# Second constructor: distribution of the maximum of independent
# random variables given their distributions
# Third constructor: distribution of the maximum of independent
# identically distributed random variables
R = ot.CorrelationMatrix(3)
R[0, 1] = 0.5
coll = [
    ot.MaximumDistribution(ot.Normal(5)),
    ot.MaximumDistribution([ot.Normal()] * 5),
    ot.MaximumDistribution(ot.Normal(), 5),
    ot.MaximumDistribution(ot.Normal([0.0] * 3, [1.0] * 3, R)),
]
for index, distribution in enumerate(coll):
    print("Distribution ", distribution)

    # Is this distribution elliptical ?
    print("Elliptical = ", distribution.isElliptical())

    # Is this distribution continuous ?
    print("Continuous = ", distribution.isContinuous())

    # Confidence regions
    (
        interval,
        threshold,
    ) = distribution.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
    if index < 3:
        ott.assert_almost_equal(interval.getLowerBound(), [-0.11278], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [2.50221], 1e-4, 0.0)
    else:
        ott.assert_almost_equal(interval.getLowerBound(), [-0.770165], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [2.33536], 1e-4, 0.0)
    print("threshold=", threshold)

    levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
    print("Minimum volume level set=", levelSet)
    print("beta=%.5g" % beta)

    (
        interval,
        beta,
    ) = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
    print("Bilateral confidence interval=", interval)
    print("beta=%.5g" % beta)

    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, False
    )
    if index < 3:
        ott.assert_almost_equal(interval.getLowerBound(), [-7.65063], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [2.31868], 1e-4, 0.0)
    else:
        ott.assert_almost_equal(interval.getLowerBound(), [-7.65063], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [2.10094], 1e-4, 0.0)
    print("beta=%.5g" % beta)

    (
        interval,
        beta,
    ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
        0.95, True
    )
    if index < 3:
        ott.assert_almost_equal(interval.getLowerBound(), [0.123843], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [7.65063], 1e-4, 0.0)
    else:
        ott.assert_almost_equal(interval.getLowerBound(), [-0.502767], 1e-4, 0.0)
        ott.assert_almost_equal(interval.getUpperBound(), [7.65063], 1e-4, 0.0)
    print("beta=%.5g" % beta)

    spearman = distribution.getSpearmanCorrelation()
    print("spearman=", spearman)
    kendall = distribution.getKendallTau()
    print("kendall=", kendall)
    print("Standard representative=", distribution.getStandardRepresentative())

    ot.Log.Show(ot.Log.TRACE)
    ot.RandomGenerator.SetSeed(1234)
    validation = ott.DistributionValidation(distribution)
    validation.skipEntropy()  # slow
    validation.skipMinimumVolumeLevelSet()  # slow
    validation.run()

# Issue #1643
coll = [ot.Uniform(), ot.Normal()]
distribution = ot.MaximumDistribution(coll)
print("%.5g" % distribution.computePDF([1.1]))



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import math as m

ot.TESTPREAMBLE()


#
# Physical model
#

inputFunction = ot.Description.BuildDefault(7, "X")[1:]

formulas = [
    "X1 + 2*X2 + 2*X3 + X4 - 5*X5 - 5*X6 +0.001*(sin(100*X1)+sin(100*X2)+sin(100*X3)+sin(100*X4)+sin(100*X5)+sin(100*X6))"
]

limitState = ot.SymbolicFunction(inputFunction, formulas)

dim = limitState.getInputDimension()
print(dim)

#
# Probabilistic model
#

mean = ot.Point(dim, 0.0)
mean[0] = 120.0
mean[1] = 120.0
mean[2] = 120.0
mean[3] = 120.0
mean[4] = 50.0
mean[5] = 40.0

sigma = ot.Point(dim, 0.0)
sigma[0] = 12.0
sigma[1] = 12.0
sigma[2] = 12.0
sigma[3] = 12.0
sigma[4] = 15.0
sigma[5] = 12.0

BorneInf = 0.0

component = ot.Description(1)

aCollection = ot.DistributionCollection()
for i in range(len(mean)):
    marginal = ot.LogNormalMuSigma(mean[i], sigma[i], BorneInf).getDistribution()
    aCollection.add(marginal)

# Create a copula : IndependentCopula (pas de correlation
aCopula = ot.IndependentCopula(aCollection.getSize())

# Instantiate one distribution object
myDistribution = ot.JointDistribution(aCollection, aCopula)

Covariance = myDistribution.getCovariance()

#
# limit state
#

vect = ot.RandomVector(myDistribution)

output = ot.CompositeRandomVector(limitState, vect)

myEvent = ot.ThresholdEvent(output, ot.Less(), 0.0)

#
# Calculs
#
#
# FORM/SORM Cobyla
myCobyla = ot.Cobyla()
myCobyla.setMaximumCallsNumber(1000 * dim)
myCobyla.setMaximumAbsoluteError(1.0e-4)
myCobyla.setMaximumRelativeError(1.0e-4)
myCobyla.setMaximumResidualError(1.0e-4)
myCobyla.setMaximumConstraintError(3.0e-3)
myCobyla.setStartingPoint(myDistribution.getMean())

myAlgoC = ot.FORM(myCobyla, myEvent)

myAlgoC.run()

resultC = ot.FORMResult(myAlgoC.getResult())

#
# FORM/SORM Abdo Rackwitz
myAbdoRackwitz = ot.AbdoRackwitz()
myAbdoRackwitz.setMaximumCallsNumber(1000 * dim)
myAbdoRackwitz.setMaximumAbsoluteError(1.0e-6)
myAbdoRackwitz.setMaximumRelativeError(1.0e-6)
myAbdoRackwitz.setMaximumResidualError(1.0e-6)
myAbdoRackwitz.setMaximumConstraintError(3.0e-3)
myAbdoRackwitz.setStartingPoint(myDistribution.getMean())

myAlgoAR = ot.FORM(myAbdoRackwitz, myEvent)

myAlgoAR.run()

resultAR = ot.FORMResult(myAlgoAR.getResult())

#
# Monte Carlo
CoV_MC = 0.5
experiment = ot.MonteCarloExperiment()
myMC = ot.ProbabilitySimulationAlgorithm(myEvent, experiment)
myMC.setMaximumOuterSampling(1000)
myMC.setBlockSize(100)
myMC.setMaximumCoefficientOfVariation(CoV_MC)
myMC.run()
result = myMC.getResult()
ott.assert_almost_equal(result.getProbabilityEstimate(), 0.0133333)

#
# LHS
CoV_LHS = 0.1
experiment = ot.LHSExperiment()
experiment.setAlwaysShuffle(True)
myLHS = ot.ProbabilitySimulationAlgorithm(myEvent, experiment)
myLHS.setMaximumOuterSampling(1000)
myLHS.setBlockSize(10)
myLHS.setMaximumCoefficientOfVariation(CoV_LHS)
myLHS.run()

#
# Directional Sampling
CoV_DS = 0.1
myDS1 = ot.DirectionalSampling(
    myEvent,
    ot.RootStrategy(ot.RiskyAndFast()),
    ot.SamplingStrategy(ot.RandomDirection()),
)
myDS1.setMaximumOuterSampling(1000)
myDS1.setBlockSize(10)
myDS1.setMaximumCoefficientOfVariation(CoV_DS)
myDS1.run()

myDS2 = ot.DirectionalSampling(
    myEvent, ot.RootStrategy(ot.MediumSafe()), ot.SamplingStrategy(ot.RandomDirection())
)
myDS2.setMaximumOuterSampling(1000)
myDS2.setBlockSize(10)
myDS2.setMaximumCoefficientOfVariation(CoV_DS)
myDS2.run()

myDS3 = ot.DirectionalSampling(
    myEvent,
    ot.RootStrategy(ot.SafeAndSlow()),
    ot.SamplingStrategy(ot.RandomDirection()),
)
myDS3.setMaximumOuterSampling(1000)
myDS3.setBlockSize(10)
myDS3.setMaximumCoefficientOfVariation(CoV_DS)
myDS3.run()

#
# Importance Sampling avec Standard Event
meanSE = ot.Point(dim, 0.0)
for i in range(resultAR.getStandardSpaceDesignPoint().getDimension()):
    meanSE[i] = resultAR.getStandardSpaceDesignPoint()[i]
sigmaSE = ot.Point(dim, 1.0)
CorrSE = ot.IdentityMatrix(dim)
myImportanceSE = ot.Normal(meanSE, sigmaSE, CorrSE)

myStandardEvent = ot.StandardEvent(myEvent)

experiment = ot.ImportanceSamplingExperiment(myImportanceSE)
myISS = ot.ProbabilitySimulationAlgorithm(myStandardEvent)
myISS.setMaximumOuterSampling(1000)
myISS.setBlockSize(10)
myISS.setMaximumCoefficientOfVariation(0.1)
myISS.run()

# Importance Sampling avec Event
meanE = ot.Point(dim, 0.0)
for i in range(resultC.getPhysicalSpaceDesignPoint().getDimension()):
    meanE[i] = resultC.getPhysicalSpaceDesignPoint()[i]

sigmaE = ot.Point(dim, 0.0)
for i in range(resultC.getPhysicalSpaceDesignPoint().getDimension()):
    sigmaE[i] = m.sqrt(Covariance[i, i])

CorrE = ot.IdentityMatrix(dim)

myImportanceE = ot.Normal(meanE, sigmaE, CorrE)

experiment = ot.ImportanceSamplingExperiment(myImportanceE)
myIS = ot.ProbabilitySimulationAlgorithm(myEvent)
myIS.setMaximumOuterSampling(1000)
myIS.setBlockSize(10)
myIS.setMaximumCoefficientOfVariation(0.1)
myIS.run()



import openturns as ot

R = ot.CorrelationMatrix(3)
R[0, 1] = 0.1
R[0, 2] = 0.2
R[1, 2] = 0.3
d = ot.Normal([0.0] * 3, [1.5, 2.0, 3.0], R)
# Test conditional methods
y = [1.2, 2.3]
x = 1.5
pc = ot.PointConditionalDistribution(d, [0, 1], y)
cdf1 = pc.computeCDF(x)
print(cdf1, pc.computeQuantile(cdf1)[0])
cdf2 = d.computeConditionalCDF(x, y)
print(cdf2, d.computeConditionalQuantile(cdf2, y))

# Test sequential conditional methods
# First, check if conditionalCDF and conditionalQuantile are inverse
xSeq = ot.Point([2.1, 2.2, 2.3])
cdfSeq = d.computeSequentialConditionalCDF(xSeq)
qSeq = d.computeSequentialConditionalQuantile(cdfSeq)
print("xSeq=", xSeq, "cdfSeq=", cdfSeq, "qSeq=", qSeq)
uSeq = ot.Point([0.1, 0.2, 0.3])
qSeq = d.computeSequentialConditionalQuantile(uSeq)
cdfSeq = d.computeSequentialConditionalCDF(qSeq)
print("uSeq=", uSeq, "qSeq=", qSeq, "cdfSeq=", cdfSeq)
# Second, check if the sequential versions match the pointwise methods
print("#" * 50)
print("computeConditionalCDF vs computeSequentialConditionalCDF")
for i in range(3):
    y = qSeq[0:i]
    x = qSeq[i]
    ind = list(range(i + 1))
    indred = list(range(i))
    pc = ot.PointConditionalDistribution(d.getMarginal(ind), indred, y)
    print(d.computeConditionalCDF(x, y), cdfSeq[i], pc.computeCDF(x))
print("#" * 50)
print("computeConditionalQuantile vs computeSequentialConditionalQuantile")
for i in range(3):
    y = qSeq[0:i]
    p = cdfSeq[i]
    ind = list(range(i + 1))
    indred = list(range(i))
    pc = ot.PointConditionalDistribution(d.getMarginal(ind), indred, y)
    print(d.computeConditionalQuantile(p, y), qSeq[i], pc.computeQuantile(p)[0])
# Third, check if the sequential versions match the pointwise methods
print("#" * 50)
pdfSeq = d.computeSequentialConditionalPDF(qSeq)
print("computeConditionalPDF vs computeSequentialConditionalPDF")
for i in range(3):
    y = qSeq[0:i]
    x = qSeq[i]
    ind = list(range(i + 1))
    indred = list(range(i))
    pc = ot.PointConditionalDistribution(d.getMarginal(ind), indred, y)
    print(d.computeConditionalPDF(x, y), pdfSeq[i], pc.computePDF(x))



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


f = ot.SymbolicFunction(["x"], ["abs(sin(x))"])
a = -2.5
b = 4.5
algo = ot.GaussKronrod(100000, 1e-13, ot.GaussKronrodRule(ot.GaussKronrodRule.G11K23))

value = algo.integrate(f, ot.Interval(a, b))[0]

ai = ot.Point()
bi = ot.Point()
fi = ot.Sample()
ei = ot.Point()
error = ot.Point()
value2 = algo.integrate(f, a, b, error, ai, bi, fi, ei)[0]

ai.add(b)
g = f.draw(a, b, 512)
lower = ot.Cloud(ai, ot.Point(ai.getDimension()))
lower.setColor("magenta")
lower.setPointStyle("circle")
g.add(lower)
g.setTitle(r"GaussKronrod example: $\int_{-5/2}^{9/2}|\sin(t)|\,dt=$" + str(value))
fig = plt.figure(figsize=(8, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(g, figure=fig, axes=[axis], add_legend=False)



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

# Create the Pie
pie = ot.Pie([0.4, 0.3, 0.2, 0.1], ["a0", "a1", "a2", "a3"])
# Create an empty graph
graph = ot.Graph("Pie example")
graph.add(pie)

# Then, draw it
fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
otv.View(graph, figure=fig, axes=[axis], add_legend=False)
axis.set_xlim(auto=True)



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott
import sys


def progress(percent):
    sys.stderr.write("-- progress=" + str(percent) + "%\n")


def stop():
    sys.stderr.write("-- stop?\n")
    return False


# List available algorithms
print(ot.Bonmin.GetAlgorithmNames())

# Definition of objective function
objectiveFunction = ot.SymbolicFunction(["x0", "x1", "x2", "x3"], ["-x0 -x1 -x2"])

# Definition of variables bounds
bounds = ot.Interval([0] * 4, [1, 1e6, 1e6, 5])

# Definition of constraints
# Constraints in OpenTURNS are defined as g(x) = 0 and h(x) >= 0
#    No equality constraint -> nothing to do
#    Inequality constraints:
h = ot.SymbolicFunction(
    ["x0", "x1", "x2", "x3"],
    ["-(x1-1/2)^2 - (x2-1/2)^2 + 1/4", "-x0 + x1", "-x0 - x2 - x3 + 2"],
)

# Definition of variables types
variablesType = [
    ot.OptimizationProblemImplementation.BINARY,
    ot.OptimizationProblemImplementation.CONTINUOUS,
    ot.OptimizationProblemImplementation.CONTINUOUS,
    ot.OptimizationProblemImplementation.INTEGER,
]

# Setting up Bonmin problem
problem = ot.OptimizationProblem(objectiveFunction)
problem.setBounds(bounds)
problem.setVariablesType(variablesType)
problem.setInequalityConstraint(h)

algo = ot.Bonmin(problem)
algo.setStartingPoint([0.0] * 4)
algo.setMaximumCallsNumber(10000)
algo.setMaximumIterationNumber(1000)
# algo.setProgressCallback(progress)
# algo.setStopCallback(stop)

for name in ot.Bonmin.GetAlgorithmNames():
    print(f"-- {name} algorithm...")
    algo.setAlgorithmName(name)
    algo.run()
    result = algo.getResult()
    x_star = result.getOptimalPoint()
    print(f"x*={x_star}")
    y_star = result.getOptimalValue()
    neval = result.getCallsNumber()
    print(f"f(x*)={y_star} neval={neval}")
    ott.assert_almost_equal(result.getOptimalPoint(), [1, 1, 0.5, 0], 1, 1e-2)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
distribution = ot.Burr(1.5, 4.5)
print("Distribution ", repr(distribution))
print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", repr(oneRealization))

# Define a point
point = ot.Point(distribution.getDimension(), 1.5)
print("Point= ", repr(point))

# Show PDF and CDF of point
# DDF = distribution.computeDDF( point )
# print "ddf     =", repr(DDF)

PDF = distribution.computePDF(point)
print("pdf     = %.12g" % PDF)

CDF = distribution.computeCDF(point)
print("cdf= %.12g" % CDF)
# CF = distribution.computeCharacteristicFunction( point[0] )
# print "characteristic function=", CF

CDFgr = distribution.computeCDFGradient(point)
print("cdf gradient     =", repr(CDFgr))

quantile = distribution.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)=", distribution.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f"
    % distribution.computeSurvivalFunction(inverseSurvival)
)
print("entropy=%.6f" % distribution.computeEntropy())

# Confidence regions
interval, threshold = distribution.computeMinimumVolumeIntervalWithMarginalProbability(
    0.95
)
ott.assert_almost_equal(interval.getLowerBound(), [0.000634766])
ott.assert_almost_equal(interval.getUpperBound(), [0.96401], 1e-3, 0.0)
print("threshold=", ot.Point(1, threshold))
levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
print("Minimum volume level set=", levelSet)
print("beta=", ot.Point(1, beta))
interval, beta = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(
    0.95
)
print("Bilateral confidence interval=", interval)
print("beta=", ot.Point(1, beta))
(
    interval,
    beta,
) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(0.95, False)
print("Unilateral confidence interval (lower tail)=", interval)
print("beta=", ot.Point(1, beta))
(
    interval,
    beta,
) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(0.95, True)
print("Unilateral confidence interval (upper tail)=", interval)
print("beta=", ot.Point(1, beta))

mean = distribution.getMean()
print("mean=", repr(mean))
standardDeviation = distribution.getStandardDeviation()
print("standard deviation=", repr(standardDeviation))
skewness = distribution.getSkewness()
print("skewness=", repr(skewness))
kurtosis = distribution.getKurtosis()
print("kurtosis=", repr(kurtosis))
covariance = distribution.getCovariance()
print("covariance=", repr(covariance))
parameters = distribution.getParametersCollection()
print("parameters=", repr(parameters))
print("Standard representative=", distribution.getStandardRepresentative())

# computeProba test with bound far away
p = distribution.computeProbability(
    ot.Interval(-ot.SpecFunc.Infinity, ot.SpecFunc.Infinity)
)
ott.assert_almost_equal(p, 1.0)

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.run()



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import chaboche_model


ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)

"""
Test the import of the ChabocheModel data class.
"""
cm = chaboche_model.ChabocheModel()

# test parameters
assert_almost_equal(cm.dim, 4, 1e-12)
assert_almost_equal(cm.trueR, 750.0e6, 1e-12)
assert_almost_equal(cm.trueC, 2750.0e6, 1e-12)
assert_almost_equal(cm.trueGamma, 10.0, 1e-12)

# test marginals means
assert_almost_equal(cm.Strain.getMean()[0], 0.035, 1e-12)

assert_almost_equal(cm.R.getMean()[0], 750.0e6, 1e-12)
assert_almost_equal(cm.C.getMean()[0], 2750.0e6, 1e-12)
assert_almost_equal(cm.Gamma.getMean()[0], 10.0, 1e-12)

# special value
X = ot.Point([1.0, 1.0, 1.0, 1.0])
assert_almost_equal(cm.model(X)[0], 1.6321, 1e-3)



#!/usr/bin/env python

import openturns as ot
import openturns.testing as ott

# Define the problems based on Rosebrock function
rosenbrock = ot.SymbolicFunction(["x1", "x2"], ["(1-x1)^2+100*(x2-x1^2)^2"])

unboundedProblem = ot.OptimizationProblem(rosenbrock)

notConstrainingBounds = ot.Interval([-5.0, -5.0], [5.0, 5.0])
constrainingBounds = ot.Interval([0.0, -2.0], [5.0, 0.5])

start = [3.0, -1.5]
unboundedPref = [1.0, 1.0]
boundedPref = [0.70856, 0.5]

for bound in [False, True]:
    for bounds in [ot.Interval(), notConstrainingBounds, constrainingBounds]:
        problem = ot.OptimizationProblem(rosenbrock)
        problem.setBounds(bounds)
        for algoName in ["cg", "bfgs", "lbfgs", "newton", "trust_region"]:
            if algoName == "trust_region" and problem.hasBounds():
                continue
            algo = ot.Dlib(problem, algoName)
            algo.setStartingPoint(start)
            algo.setMaximumIterationNumber(10000)
            algo.setMaximumCallsNumber(100000)
            algo.run()
            result = algo.getResult()
            x = result.getOptimalPoint()
            y = result.getOptimalValue()
            print(f"bounds={problem.hasBounds()} algo={algoName} x^={x} y^={y}")
            if bounds == constrainingBounds:
                ott.assert_almost_equal(x, boundedPref, 5e-2)
            else:
                ott.assert_almost_equal(x, unboundedPref, 5e-2)



"""
Control algorithm termination
=============================
"""

# %%
# In this examples we are going to expose ways to control the termination of optimization and simulation algorithms using callbacks.
#

# %%
import openturns as ot


# %%
# Define an event to compute a probability
myFunction = ot.SymbolicFunction(["E", "F", "L", "I"], ["-F*L^3/(3.0*E*I)"])
dim = myFunction.getInputDimension()
mean = [50.0, 1.0, 10.0, 5.0]
sigma = [1.0] * dim
R = ot.IdentityMatrix(dim)
myDistribution = ot.Normal(mean, sigma, R)
vect = ot.RandomVector(myDistribution)
output = ot.CompositeRandomVector(myFunction, vect)
myEvent = ot.ThresholdEvent(output, ot.Less(), -3.0)

# %%
# **Stop a FORM algorithm using a calls number limit**
#
# A FORM algorithm termination can be controlled by the maximum number of iterations
#
# of its underlying optimization solver, but not directly by a maximum number of evaluations.

# %%
# Create the optimization algorithm
myCobyla = ot.Cobyla()
myCobyla.setMaximumCallsNumber(400)
myCobyla.setMaximumAbsoluteError(1.0e-10)
myCobyla.setMaximumRelativeError(1.0e-10)
myCobyla.setMaximumResidualError(1.0e-10)
myCobyla.setMaximumConstraintError(1.0e-10)


# %%
# Define the stopping criterion
def stop():
    return myFunction.getCallsNumber() > 100


myCobyla.setStopCallback(stop)

# %%
# Run FORM
myCobyla.setStartingPoint(mean)
algo = ot.FORM(myCobyla, myEvent)
algo.run()
result = algo.getResult()
print("event probability:", result.getEventProbability())
print("calls number:", myFunction.getCallsNumber())

# %%
# **Stop a simulation algorithm using a time limit**
#
# Here we will create a callback to not exceed a specified simulation time.

# %%
# Create simulation
experiment = ot.MonteCarloExperiment()
algo = ot.ProbabilitySimulationAlgorithm(myEvent, experiment)
algo.setMaximumOuterSampling(1000000)
algo.setMaximumCoefficientOfVariation(-1.0)

# %%
# Define the stopping criterion
algo.setMaximumTimeDuration(0.01)

# %%
# Run the algorithm
algo.run()
result = algo.getResult()
pf = result.getProbabilityEstimate()
nCalls = myFunction.getCallsNumber()



#! /usr/bin/env python

import openturns.coupling_tools as ct
import os
import time
import sys
import openturns.testing as ott
import subprocess

wanted_lines = "# ooo\nE=@E\nE1=@E\nFE1=@F#oo\nZ=@Z@Z\n# ooo\n"
semi_parsed = "# ooo\nE=2\nE1=2\nFE1=@F#oo\nZ=@Z@Z\n# ooo\n"
parsed = "# ooo\nE=1.6\nE1=1.6\nFE1=5#oo\nZ=66\n# ooo\n"

# how many Mo for perf file
howbig = 128
max_time = howbig // 5
# enable it manually, see #600
max_time = sys.maxsize


def create_template():
    template_name = "template.in"
    with open(template_name, "wb") as f:
        f.write(wanted_lines.encode())
    return template_name


def create_big_template():
    template_name = "template_big.in"

    print(("create template file of " + str(howbig) + "Mo"))
    with open(template_name, "wb") as f:
        f.write(wanted_lines.encode())
        for i in range(howbig):
            for i in range(1024):
                # line of 1024 octets
                f.write(b"u" * 1024)
                f.write(b"\n")
        f.write(b"# ooo\n")

    return template_name


def remove_file(filename, quiet=False):
    if quiet:
        try:
            os.remove(filename)
        except Exception:
            pass
    else:
        os.remove(filename)


def check_outfile(filename, wanted_result):
    """wanted_result: a string"""
    is_ok = True
    with open(filename) as f:
        for wanted_line, result_line in zip(wanted_result.splitlines(True), f):
            if wanted_line != result_line:
                print(
                    (
                        "Aaaaarg, result is not what we wanted (result:"
                        + result_line
                        + ", should be:"
                        + wanted_line.decode()
                        + ")"
                    )
                )
                is_ok = False

    if is_ok:
        print(("check " + filename + ": ok"))
    else:
        exit(1)
    # return is_ok


def check_replace():
    print(("=== " + sys._getframe().f_code.co_name))
    print("= check replace std")
    template = create_template()
    template_out = template + ".replaced"
    ct.replace(infile=template, outfile=template_out, tokens=["@E"], values=[2])
    check_outfile(template_out, semi_parsed)
    remove_file(template_out)
    remove_file(template)

    print("= check replace more vars")
    template = create_template()
    ct.replace(
        infile=template,
        outfile=template_out,
        tokens=["@E", "@F", "@Z"],
        values=[1.6, 5, 6],
    )
    check_outfile(template_out, parsed)
    remove_file(template_out)
    remove_file(template)

    print("= check replace inplace")
    template = create_template()
    ct.replace(
        infile=template, outfile=template, tokens=["@E", "@F", "@Z"], values=[1.6, 5, 6]
    )
    check_outfile(template, parsed)
    remove_file(template)

    print("= check replace inplace with None")
    template = create_template()
    ct.replace(
        infile=template, outfile=None, tokens=["@E", "@F", "@Z"], values=[1.6, 5, 6]
    )
    check_outfile(template, parsed)
    remove_file(template)

    print("= check replace big template")
    start_time = time.time()
    template = create_big_template()
    sys.stderr.write(
        "big template created in : " + str(time.time() - start_time) + "s\n"
    )

    template_out = template + ".replaced"
    start_time = time.time()
    ct.replace(infile=template, outfile=template_out, tokens=["@E"], values=[2])
    time_to_parse = str(int(time.time() - start_time))
    check_outfile(template_out, semi_parsed)
    remove_file(template_out)
    remove_file(template)
    sys.stderr.write("parsed template in: " + time_to_parse + "s\n")
    # parsed template=3G -> 25s on bx (ssd, core i7@2.5GHz)
    if int(time_to_parse) > max_time:
        print(
            ("time to get token took too long (should be " + str(max_time) + "s max)")
        )
        exit(1)
    else:
        print("check replace big template: ok")


def create_results(tokens, values=None, big=False):
    filename = "results.out"

    with open(filename, "wb") as f:

        if big:
            print(("create file of " + str(howbig) + "Mo"))

            for i in range(howbig):
                for i in range(1024):
                    # line of 1024 octets
                    f.write(b"u" * 1024)
                    f.write(b"\n")
            f.write(b"# ooo\n")

        if values is None:
            f.write(tokens.encode())
        else:
            n = 0
            for t, v in zip(tokens, values):
                f.write((t + str(v)).encode())
                # go to next line sometimes
                if n % 3 == 0:
                    f.write(b"\n")
                n += 1

    return filename


def check_results(ok_values, values):
    if ok_values == values:
        print("ok")
    else:
        print(("Error: found: " + str(values) + " should be: " + str(ok_values)))
        exit(1)


def check_get_line_col():
    print(("=== " + sys._getframe().f_code.co_name))
    content = """01 02 03 04 05 06 07 08 09
11 12 13 14 15 16 17 18 19
21 22 23 24 25 26 27 28 29
31 32 33 34 35 36 37 38 39
"""
    result_file = create_results(content)

    value = 1
    result = ct.get_line_col(result_file)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 3
    result = ct.get_line_col(result_file, skip_col=2)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 11
    result = ct.get_line_col(result_file, 1)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 16
    result = ct.get_line_col(result_file, 1, 5)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 9
    result = ct.get_line_col(result_file, skip_col=-1)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 17
    result = ct.get_line_col(result_file, 1, -3)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 31
    result = ct.get_line_col(result_file, -1)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 14
    result = ct.get_line_col(result_file, -3, -6)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 3
    result = ct.get_line_col(result_file, seek=6)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 23
    result = ct.get_line_col(result_file, skip_line=1, skip_col=2, seek=30)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))
    value = 31
    result = ct.get_line_col(result_file, skip_line=-1, seek=-100)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))

    # ct.debug = True
    value = 21
    just_before_line_ret = 80
    sys.stderr.write(
        "char at pos "
        + str(just_before_line_ret)
        + ":->"
        + content[just_before_line_ret]
        + "<-\n"
    )
    result = ct.get_line_col(result_file, skip_line=-1, seek=-just_before_line_ret)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))

    value = 21
    just_after_line_ret = just_before_line_ret + 1
    sys.stderr.write(
        "char at pos "
        + str(just_after_line_ret)
        + ":->"
        + content[just_after_line_ret]
        + "<-\n"
    )
    result = ct.get_line_col(result_file, skip_line=-2, seek=-just_after_line_ret)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))

    try:
        result = ct.get_line_col(result_file, skip_line=4)
    except Exception:
        pass
    else:
        raise Exception("! should have fail !")

    try:
        result = ct.get_line_col(result_file, skip_line=-5)
    except Exception:
        pass
    else:
        raise Exception("! should have fail !")

    os.remove(result_file)

    # test with a last empty line
    content = """01 02 03 04 05 06 07 08 09
11 12 13 14 15 16 17 18 19

"""
    result_file = create_results(content)

    value = 19
    result = ct.get_line_col(result_file, skip_line=-2, skip_col=-1)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))

    value = 1
    result = ct.get_line_col(result_file, skip_line=-3)
    if value != result:
        raise Exception("! got " + str(result) + " instead of " + str(value))

    try:
        result = ct.get_line_col(result_file, skip_line=-1)
    except Exception:
        pass
    else:
        raise Exception("! should have fail !")

    os.remove(result_file)

    print("ok")


def check_get():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = ["@Y1=", "@Y2="]
    values = [5.4, 6.5]
    result_file = create_results(tokens, values)
    results = ct.get(filename=result_file, tokens=tokens)
    check_results(values, results)
    os.remove(result_file)

    # Y = ct.get(outfile="output.py", tokens=["@Y1=", "@Y2="],
    #                       occurrence=[0, 0], col=[0, 3], line=[0, 2])


def check_get_token_escaped():
    print(("=== " + sys._getframe().f_code.co_name))
    result_file = "results.out"
    with open(result_file, "wb") as f:
        f.write(b"FOO (BAR\n")
        f.write(b"1.0 1.1 1.2 1.3\n")
        f.write(b"2.0 2.1 2.2 2.3\n")
    results = ct.get(
        filename=result_file, tokens=[r"FOO \(BAR"], skip_lines=[1], skip_cols=[1]
    )
    check_results([1.1], results)
    os.remove(result_file)


def check_get_regex():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = ["@E=", "02=", " 01 = "]
    values = [-9.55555e5, 8, 5.4]
    result_file = create_results(tokens, values)
    results = ct.get_regex(
        filename=result_file, patterns=[r"@E=(\R)", r"02\s*=\s*(\I)\s*", r"01 =\s*(\R)"]
    )
    check_results(values, results)
    remove_file(result_file)


def check_get_regex_perf():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = ["@E=", "02=", " 01 = "]
    values = [-9.55555e5, 8, 5.4]
    start_time = time.time()
    result_file = create_results(tokens, values, big=True)
    sys.stderr.write("big file created in : " + str(time.time() - start_time) + "s\n")
    start_time = time.time()
    results = ct.get_regex(
        filename=result_file, patterns=[r"@E=(\R)", r"02\s*=\s*(\I)\s*", r"01 =\s*(\R)"]
    )
    time_to_parse = str(int(time.time() - start_time))
    check_results(values, results)
    remove_file(result_file)
    # get file=3G -> 16s on bx (ssd, core i7@2.5GHz)
    sys.stderr.write("get regex in file in: " + time_to_parse + "s\n")
    if int(time_to_parse) > max_time:
        print(
            ("time to get token took too long (should be " + str(max_time) + "s max)")
        )
        exit(1)
    else:
        print("get regex in file: ok")


def check_get_tokens():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = ["@E=", " pp", ",virg", " normal="]
    values = [-9.55555e6, 56.666, -12345678912.2, 0]
    result_file = create_results(tokens, values)
    results = ct.get(filename=result_file, tokens=tokens)
    check_results(values, results)
    remove_file(result_file)


def check_get_tokens_skip():
    print(("=== " + sys._getframe().f_code.co_name))
    content = "@E=99 @E=-9.55555E6 pp88 pp 56.666,virg-12345678912.2 normal=0"
    values = [-9.55555e6, 56.666, -12345678912.2, 0]
    result_file = create_results(content)
    results = ct.get(
        filename=result_file,
        tokens=["@E=", "pp", ",virg", "normal="],
        skip_tokens=[1, -1, 0, 0],
    )
    check_results(values, results)
    remove_file(result_file)

    print(("=== " + sys._getframe().f_code.co_name + "2"))
    tokens = (
        "@E=99 @E=7899 pp88 pp pp\n"
        "pp999 pp56.666E-9pp,virg-12345678912.2 uu88 uuuu\n"
        "uu999uu\n"
    )
    values = [99, 56.666e-9, -12345678912.2, 999]
    result_file = create_results(tokens)
    results = ct.get(
        filename=result_file,
        tokens=["@E=", "pp", ",virg", "uu"],
        skip_tokens=[0, 4, -1, 3],
    )
    check_results(values, results)
    remove_file(result_file)


def check_get_array():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = (
        "11.0E-9 22.0crap 33.0 44.0 55.0\n"
        "11.1 22.1 33.1 44.1\n"
        "11.2 22.2 33.2 44.2"
    )
    values = [11.0e-9, 22.0, 55.0, 11.1, 33.2, 22.2, 33.2]
    result_file = create_results(tokens)
    results = ct.get(
        filename=result_file,
        skip_lines=[0, 0, 0, 1, 2, 2, -1],
        skip_cols=[0, 1, -1, 0, 2, 1, -2],
    )
    check_results(values, results)
    remove_file(result_file)


def check_get_tokens_line_col():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = (
        "11.0E-9 22.0crap 33.0 44.0 55.0\n"
        "11.1 22.1 33.1 44.1 middle\n"
        "11.2 22.2 33.2 44.2\n"
        "@E=1111.1E11 666"
    )
    values = [1111.1e11, 22.0, 33.1, 666, 33.2, 44.1, 55.0]
    result_file = create_results(tokens)
    results = ct.get(
        filename=result_file,
        tokens=["@E=", None, "@E=", "@E=", "middle", "middle", "middle"],
        skip_lines=[0, 0, -2, 0, 1, 0, -1],
        skip_cols=[0, 1, 2, 1, -2, -1, -1],
    )
    check_results(values, results)
    remove_file(result_file)


def check_get_tokens_skip_line_col():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = (
        "11.0E-9 22.0crap 33.0 44.0 55.0\n"
        "11.1 22.1 33.1 44.1 middle\n"
        "11.2 22.2 middle 33.2 44.2\n"
        "@E=1111.1E11 666\n"
        "@E=999 8888 @E=95\n"
    )
    values = [1111.1e11, 33.2, 666, 8888, 8888, 666]
    result_file = create_results(tokens)
    results = ct.get(
        filename=result_file,
        tokens=["@E=", "middle", "middle", "@E=", "@E=", "@E="],
        skip_tokens=[0, 1, -1, -2, -1, -2],
        skip_lines=[0, 0, 1, 0, 0, -1],
        skip_cols=[0, 0, 1, 1, -1, -1],
    )
    check_results(values, results)
    remove_file(result_file)


def check_get_tokens_perf():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = ["@E=", " pp", ",virg", " normal="]
    values = [-9.55555e6, 56.666, -12345678912.2, 0]
    tokens = ["@E="]
    values = [-9.55555e6]
    start_time = time.time()
    result_file = create_results(tokens, values, big=True)
    sys.stderr.write("big file created in : " + str(time.time() - start_time) + "s\n")
    start_time = time.time()
    results = ct.get(filename=result_file, tokens=tokens)
    time_to_parse = str(int(time.time() - start_time))
    check_results(values, results)
    remove_file(result_file)
    # get file=3G -> 18s on bx (ssd, core i7@2.5GHz)
    sys.stderr.write("get token in file in: " + time_to_parse + "s\n")
    if int(time_to_parse) > max_time:
        print(
            ("time to get token took too long (should be " + str(max_time) + "s max)")
        )
        exit(1)
    else:
        print("get tokens: ok")


def check_get_tokens_skip_perf():
    print(("=== " + sys._getframe().f_code.co_name))
    content = "@E=99 @E=-9.55555E6 pp88 pp 56.666,virg-12345678912.2 normal=0"
    values = [-9.55555e6, 56.666, -12345678912.2, 0]
    content = "@E=99 @E=-9.55555E6 pp88 pp 56.666,virg-12345678912.2 normal=0"
    values = [-9.55555e6]
    start_time = time.time()
    result_file = create_results(content, big=True)
    sys.stderr.write("big file created in : " + str(time.time() - start_time) + "s\n")
    start_time = time.time()
    results = ct.get(filename=result_file, tokens=["@E="], skip_tokens=[-1])
    time_to_parse = str(int(time.time() - start_time))
    check_results(values, results)
    remove_file(result_file)
    # get file=3G -> 21s on bx (ssd, core i7@2.5GHz)
    sys.stderr.write("get token skip in file in: " + time_to_parse + "s\n")
    if int(time_to_parse) > max_time:
        print(
            ("time to get token took too long (should be " + str(max_time) + "s max)")
        )
        exit(1)
    else:
        print("get tokens skip: ok")


def check_get_line_col_perf():
    print(("=== " + sys._getframe().f_code.co_name))
    tokens = (
        "11.0E-9 22.0 33.0 44.0 55.0\n"
        "11.1 22.1 33.1 44.1\n"
        "11.2 22.2 33.2 44.2"
        "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
        "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
        "\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n"
    )
    values = [22.0]
    start_time = time.time()
    result_file = create_results(tokens, big=True)
    sys.stderr.write("big file created in : " + str(time.time() - start_time) + "s\n")
    start_time = time.time()
    results = ct.get(filename=result_file, skip_lines=[-92], skip_cols=[1])
    time_to_parse = str(int(time.time() - start_time))
    check_results(values, results)
    remove_file(result_file)
    # get file=3G -> 15s on bx (ssd, core i7@2.5GHz)
    sys.stderr.write("get token skip line col in file in: " + time_to_parse + "s\n")
    if int(time_to_parse) > max_time:
        print(
            ("time to get token took too long (should be " + str(max_time) + "s max)")
        )
        exit(1)
    else:
        print("get line col: ok")


def check_execute():
    print(("=== " + sys._getframe().f_code.co_name))
    # ensure previous print is print before following command output
    sys.stdout.flush()

    if sys.platform.startswith("win"):
        ct.execute("cmd.exe /c echo hello")
    else:
        ct.execute("/bin/echo hello")

    ct.execute("echo hi", shell=True)

    if sys.platform.startswith("win"):
        ct.execute("echo hi", shell=True, hide_win=False)
    else:
        ct.execute("echo hi", shell=True, executable="/bin/sh")

    cp = ct.execute("echo hello", shell=True, capture_output=True)
    if cp.returncode != 0 or not cp.stdout.decode().startswith("hello"):
        raise Exception("ct.execute error!")

    cp = ct.execute("echo hello", shell=True, capture_output=True)
    if (
        cp.returncode != 0
        or not cp.stdout.decode().startswith("hello")
        or len(cp.stderr) > 0
    ):
        raise Exception("ct.execute error!")

    # we expect a subclass of CalledProcessError with the error stream in the exception message
    with ott.assert_raises(subprocess.CalledProcessError):
        cp = ct.execute(
            sys.executable + " zebuebceb745az4f801m", shell=True, capture_output=True
        )

    print("execute ok")


def check_get_value():
    with open("result.txt", "w") as f:
        f.write("name,X0,X1,X2,X3,X4,X5\n")
        f.write("val,0.125,1.1,2.3,3.5,4.7,5.99\n")
    ott.assert_almost_equal(
        ct.get_value("result.txt", skip_line=1, skip_col=3, col_sep=","), 2.3
    )
    ott.assert_almost_equal(
        ct.get_value("result.txt", skip_line=1, skip_col=-1, col_sep=","), 5.99
    )
    os.remove("result.txt")


def check_get_value_token():
    with open("result.txt", "w") as f:
        f.write("Price [£] : 42\n")
        f.write("Temperature [°C] : 12.34\n")
    tvalue = ct.get_value("result.txt", token="C] :")
    ott.assert_almost_equal(tvalue, 12.34)
    os.remove("result.txt")


check_execute()
check_replace()
check_get_token_escaped()
check_get_regex()
check_get_regex_perf()
check_get_line_col()
check_get_tokens()
check_get_tokens_skip()
check_get_array()
check_get_tokens_line_col()
check_get_tokens_skip_line_col()
check_get_tokens_perf()
check_get_tokens_skip_perf()
check_get_line_col_perf()
check_get_value()
check_get_value_token()



"""
Over-fitting and model selection
================================
"""

# %%

# %%
# Introduction
# ------------
#
# In this notebook, we present the problem of over-fitting a model to data.
# We consider noisy observations of the `sine` function.
# We estimate the coefficients of the univariate polynomial based on linear
# least squares and show that, when the degree of the polynomial becomes too
# large, the overall prediction quality decreases.
#
# This shows why and how model selection can come into play in order to select
# the degree of the polynomial: there is a trade-off between fitting the
# data and preserving the quality of future predictions.
# In this example, we use cross validation as a model selection method.

# %%
# References
# ----------
#
# * Bishop Christopher M., 1995, Neural networks for pattern recognition. Figure 1.4, page 7
#

# %%
# Compute the data
# ----------------
#
# In this section, we generate noisy observations from the `sine` function.

# %%
import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


# %%
# We define the function that we are going to approximate.
g = ot.SymbolicFunction(["x"], ["sin(2*pi_*x)"])

# %%
graph = ot.Graph("Polynomial curve fitting", "x", "y")
graph.setLegendPosition("upper right")
# The "unknown" function
curve = g.draw(0, 1)
curve.setLegends(['"Unknown" function'])
graph.add(curve)
view = otv.View(graph)


# %%
# This seems a smooth function to approximate with polynomials.


# %%
def linearSample(xmin, xmax, npoints):
    """Returns a sample created from a regular grid
    from xmin to xmax with npoints points."""
    step = (xmax - xmin) / (npoints - 1)
    rg = ot.RegularGrid(xmin, step, npoints)
    vertices = rg.getVertices()
    return vertices


# %%
# We consider 10 observation points in the interval [0,1].

# %%
n_train = 10
x_train = linearSample(0, 1, n_train)

# %%
# Assume that the observations are noisy and that the noise follows a Normal distribution with zero mean and small standard deviation.

# %%
noise = ot.Normal(0, 0.1)
noiseSample = noise.getSample(n_train)

# %%
# The following computes the observation as the sum of the function value and of the noise.
# The couple (`x_train` , `y_train`) is the training set: it is used to compute the coefficients of the polynomial model.

# %%
y_train = g(x_train) + noiseSample

# %%
graph = ot.Graph("Polynomial curve fitting", "x", "y")
graph.setLegendPosition("upper right")
# The "unknown" function
curve = g.draw(0, 1)
curve.setLegends(['"Unknown" function'])
graph.add(curve)
# Training set
cloud = ot.Cloud(x_train, y_train)
cloud.setPointStyle("circle")
cloud.setLegend("Observations")
graph.add(cloud)
view = otv.View(graph)

# %%
# Compute the coefficients of the polynomial decomposition
# --------------------------------------------------------
#
# Let :math:`\vect{y} \in \Rset^n` be a vector of observations.
# The polynomial model is
#
# .. math::
#    P(x) = \beta_0 + \beta_1 x + ... + \beta_p x^p,
#
# for any :math:`x \in \Rset`, where :math:`p` is the polynomial degree and :math:`\vect{\beta} \in \Rset^{p+1}` is the vector of the coefficients of the model.
# Let :math:`\sampleSize` be the training sample size and let :math:`x_1,...,x_\sampleSize \in \Rset` be the abscissas of the training set.
# The design matrix :math:`\mat{X} \in \Rset^{n \times (p+1)}` is
#
# .. math::
#    x_{i,j} = x^j_i,
#
# for :math:`i=1,...,n` and :math:`j=0,...,p`.
# The least squares solution is:
#
# .. math::
#    \beta^\star = \argmin_{\vect{\beta} \in \Rset^{p+1}} \| \mat{X} \vect{\beta} - \vect{y}\|_2^2.
#

# %%
# In order to approximate the function with polynomials up to degree 4,
# we create a list of strings containing the associated monomials.
# We do not include a constant in the polynomial basis, as this constant term
# is automatically included in the model by the :class:`~openturns.LinearLeastSquares` class.
# We perform the loop from 1 up to `total_degree` (but the `range` function takes `total_degree + 1` as its second input argument).

# %%
total_degree = 4
polynomialCollection = ["x^%d" % (degree) for degree in range(1, total_degree + 1)]
polynomialCollection

# %%
# Given the list of strings, we create a symbolic function which computes the values of the monomials.

# %%
basis = ot.SymbolicFunction(["x"], polynomialCollection)
basis

# %%
designMatrix = basis(x_train)
designMatrix

# %%
myLeastSquares = ot.LinearLeastSquares(designMatrix, y_train)
myLeastSquares.run()

# %%
responseSurface = myLeastSquares.getMetaModel()

# %%
# The test set
# ------------

# %%
# The couple (`x_test` , `y_test`) is the test set: it is used to assess the quality of the polynomial model with points that were not used for training.

# %%
n_test = 50
x_test = linearSample(0, 1, n_test)
y_test = responseSurface(basis(x_test))

# %%
graph = ot.Graph("Polynomial curve fitting", "x", "y")
graph.setLegendPosition("upper right")
# The "unknown" function
curve = g.draw(0, 1)
curve.setLegends(['"Unknown" function'])
graph.add(curve)
# Training set
cloud = ot.Cloud(x_train, y_train)
cloud.setPointStyle("circle")
cloud.setLegend("Observations")
graph.add(cloud)
# Predictions
curve = ot.Curve(x_test, y_test)
curve.setLegend("Polynomial Degree = %d" % (total_degree))
graph.add(curve)
view = otv.View(graph)

# %%
# Compute the residuals
# ---------------------

# %%
# For each observation in the training set, the residual is the vertical distance between the model and the observation.

# %%
# sphinx_gallery_thumbnail_number = 4
graph = ot.Graph(
    "Least squares minimizes the sum of the squares of the vertical bars",
    "x",
    "y",
)
graph.setLegendPosition("upper right")
residualsColor = ot.Drawable.BuildDefaultPalette(3)[2]
# Predictions
curve = ot.Curve(x_test, y_test)
curve.setLegend("Polynomial Degree = %d" % (total_degree))
graph.add(curve)
# Training set observations
cloud = ot.Cloud(x_train, y_train)
cloud.setPointStyle("circle")
cloud.setLegend("Observations")
graph.add(cloud)
# Errors
ypredicted_train = responseSurface(basis(x_train))
for i in range(n_train):
    curve = ot.Curve([x_train[i], x_train[i]], [y_train[i], ypredicted_train[i]])
    curve.setColor(residualsColor)
    curve.setLineWidth(2)
    if i == 0:
        curve.setLegend("Residual")
    graph.add(curve)
view = otv.View(graph)


# %%
# The least squares method minimizes the sum of the squared errors i.e. the sum of the squares of the lengths of the vertical segments.

# %%
# We gather the previous computation in two different functions. The `myPolynomialDataFitting` function computes
# the least squares solution and `myPolynomialCurveFittingGraph` plots the results.


# %%
def myPolynomialDataFitting(total_degree, x_train, y_train):
    """Computes the polynomial curve fitting
    with given total degree.
    This is for learning purposes only: please consider a serious metamodel
    for real applications, e.g. polynomial chaos or kriging."""
    polynomialCollection = ["x^%d" % (degree) for degree in range(1, total_degree + 1)]
    basis = ot.SymbolicFunction(["x"], polynomialCollection)
    designMatrix = basis(x_train)
    myLeastSquares = ot.LinearLeastSquares(designMatrix, y_train)
    myLeastSquares.run()
    responseSurface = myLeastSquares.getMetaModel()
    return responseSurface, basis


# %%
def myPolynomialCurveFittingGraph(total_degree, x_train, y_train):
    """Returns the graphics for a polynomial curve fitting
    with given total degree"""
    responseSurface, basis = myPolynomialDataFitting(total_degree, x_train, y_train)
    # Graphics
    n_test = 100
    x_test = linearSample(0, 1, n_test)
    ypredicted_test = responseSurface(basis(x_test))
    # Graphics
    graph = ot.Graph("Polynomial curve fitting", "x", "y")
    graph.setLegendPosition("upper right")
    # The "unknown" function
    curve = g.draw(0, 1)
    graph.add(curve)
    # Training set
    cloud = ot.Cloud(x_train, y_train)
    cloud.setPointStyle("circle")
    cloud.setLegend("N=%d" % (x_train.getSize()))
    graph.add(cloud)
    # Predictions
    curve = ot.Curve(x_test, ypredicted_test)
    curve.setLegend("Degree = %d" % (total_degree))
    graph.add(curve)
    return graph


# %%
# In order to see the effect of the polynomial degree, we compare the polynomial fit with degrees equal to 0 (constant), 1 (linear), 3 (cubic) and 9.

# %%
grid = ot.GridLayout(2, 2)
grid.setGraph(0, 0, myPolynomialCurveFittingGraph(0, x_train, y_train))
grid.setGraph(0, 1, myPolynomialCurveFittingGraph(1, x_train, y_train))
grid.setGraph(1, 0, myPolynomialCurveFittingGraph(3, x_train, y_train))
grid.setGraph(1, 1, myPolynomialCurveFittingGraph(9, x_train, y_train))
view = otv.View(grid, figure_kw={"figsize": (8.0, 5.0)})
plt.subplots_adjust(hspace=0.5, wspace=0.5)

# %%
# When the polynomial degree is low, the fit is satisfying.
# The polynomial is close to the observations, although there is still some residual error.
#
# However, when the polynomial degree is high, it produces large oscillations
# which significantly deviate from the true function.
# This is *overfitting*. This is a pity, given the fact that the polynomial
# *exactly* interpolates the observations: the residuals are zeroed.
#
# If the locations of the x abscissas could be changed, then the oscillations could be made smaller.
# This is the method used in Gaussian quadrature, where the nodes of interpolation are made closer on the left and right bounds.
# In our situation, we make the asssumption that these abscissas cannot be changed: the most obvious choice is to limit the degree of the polynomial.
# Another possibility is to include a regularization into the least squares solution.

# %%
# Root mean squared error
# -----------------------
#
# In order to assess the quality of the polynomial fit, we create a second dataset, the *test set* and compare the value of the polynomial with the test observations.

# %%
sqrt = ot.SymbolicFunction(["x"], ["sqrt(x)"])

# %%
# In order to see how close the model is to the observations, we compute the root mean square error.
#
# First, we create a degree 4 polynomial which fits the data.

# %%
total_degree = 4
responseSurface, basis = myPolynomialDataFitting(total_degree, x_train, y_train)


# %%
# Then we create a test set, with the same method as before.


# %%
def createDataset(n):
    x = linearSample(0, 1, n)
    noiseSample = noise.getSample(n)
    y = g(x) + noiseSample
    return x, y


# %%
n_test = 100
x_test, y_test = createDataset(n_test)

# %%
# On this test set, we evaluate the polynomial.

# %%
ypredicted_test = responseSurface(basis(x_test))

# %%
# The vector of residuals is the vector of the differences between the observations and the predictions.

# %%
residuals = y_test.asPoint() - ypredicted_test.asPoint()

# %%
# The `normSquare` method computes the square of the Euclidian norm (i.e., the 2-norm).
# We divide this by the test sample size (so as to compare the error for different sample sizes)
# and compute the square root of the result (so that the result has the same unit as `y` ).

# %%
RMSE = sqrt([residuals.normSquare() / n_test])[0]
RMSE


# %%
# The following function gathers the RMSE computation to make the experiment easier.


# %%
def computeRMSE(responseSurface, basis, x, y):
    ypredicted = responseSurface(basis(x))
    residuals = y.asPoint() - ypredicted.asPoint()
    RMSE = sqrt([residuals.normSquare() / n_test])[0]
    return RMSE


# %%
maximum_degree = 10
RMSE_train = ot.Sample(maximum_degree, 1)
RMSE_test = ot.Sample(maximum_degree, 1)
for total_degree in range(maximum_degree):
    responseSurface, basis = myPolynomialDataFitting(total_degree, x_train, y_train)
    RMSE_train[total_degree, 0] = computeRMSE(responseSurface, basis, x_train, y_train)
    RMSE_test[total_degree, 0] = computeRMSE(responseSurface, basis, x_test, y_test)

# %%
degreeSample = ot.Sample([[i] for i in range(maximum_degree)])
graph = ot.Graph("Root mean square error", "Degree", "RMSE")
graph.setLegendPosition("upper right")
# Train
cloud = ot.Curve(degreeSample, RMSE_train)
cloud.setLegend("Train")
cloud.setPointStyle("circle")
graph.add(cloud)
# Test
cloud = ot.Curve(degreeSample, RMSE_test)
cloud.setLegend("Test")
cloud.setPointStyle("circle")
graph.add(cloud)
view = otv.View(graph)

# %%
# We see that the RMSE on the train set continuously decreases, reaching zero
# when the polynomial degree is so that the number of coefficients is equal to
# the train dataset sample size. In this extreme situation, the least squares
# solution is equivalent to solving a linear system of equations: this leads to a zero residual.
#
# On the test set however, the RMSE decreases, reaches a flat region,
# then increases dramatically when the degree is equal to 9.
# Hence, limiting the polynomial degree limits overfitting.

# %%
# Increasing the training set
# ---------------------------
#
# We wonder what happens when the training dataset size is increased.

# %%
total_degree = 9
grid = ot.GridLayout(1, 2)
n_train = 11
x_train, y_train = createDataset(n_train)
grid.setGraph(0, 0, myPolynomialCurveFittingGraph(total_degree, x_train, y_train))
n_train = 100
x_train, y_train = createDataset(n_train)
grid.setGraph(0, 1, myPolynomialCurveFittingGraph(total_degree, x_train, y_train))
view = otv.View(grid, figure_kw={"figsize": (8.0, 4.0)})
plt.subplots_adjust(wspace=0.3)

# %%
# We see that the polynomial oscillates with a dataset with size 11, but does not with the larger dataset: increasing the training dataset mitigates the oscillations.
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# We create an empty Sample
sample1 = ot.Sample(0, 2)
sample1.setName("Sample1")
sample1.setDescription(("a0", "a1"))

# We populate the empty sample
point1 = ot.Point(2)
point1[0] = 10.0
point1[1] = 20.0
point2 = ot.Point(2)
point2[0] = 11.0
point2[1] = 21.0
point3 = ot.Point(2)
point3[0] = 12.0
point3[1] = 22.0
sample1.add(point1)
sample1.add(point2)
sample1.add(point3)
print("sample1=", repr(sample1))

# We get the second element of the sample
secondElement = sample1[1]
print("second element=", repr(secondElement))

# We set the third element to a valid new element
newPoint = ot.Point(2)
newPoint[0] = 1000
newPoint[1] = 2000
sample1[2] = newPoint
print("sample1=", repr(sample1))

sample2 = ot.Sample(10, point1)
print("sample2=", repr(sample2))

sample2[5] = point2
print("sample2=", repr(sample2))

try:
    # We get the tenth element of the sample
    # THIS SHOULD NORMALLY FAIL
    tenthElement = sample1[9]

    # Normally, we should never go here
    raise
except Exception:
    pass

# We try to create a sample with 5 times the same point
samePoint = ot.Point(3)
samePoint[0] = 1000.0
samePoint[1] = 2000.0
samePoint[2] = 3000.0
sample3 = ot.Sample(5, samePoint)
print("sample3=", repr(sample3))

# We append a new point to the sample...
anotherNewPoint = ot.Point(3)
anotherNewPoint[0] = -1000.0
anotherNewPoint[1] = -2000.0
anotherNewPoint[2] = -3000.0
sample3.add(anotherNewPoint)
print("sample3=", repr(sample3))

# ... then the first point again
sample3.add(samePoint)
print("sample3=", repr(sample3))

# We find the new point into the sample
pos = sample3.index(anotherNewPoint)
if pos != 5:
    print("Error in finding point into sample. Expected 5. Got ", pos)

# We erase the new point from the sample
sample3.erase(pos)
print("sample3=", repr(sample3))

# Partial copy constructor containing sample2[4:7]
sample4 = ot.Sample(sample2, 4, 8)
print("sample4=", repr(sample4))

sample5 = ot.Sample([[1.0], [3.0], [2.0]])
print("sample5=", repr(sample5))
print("sample5 (linear)=", sample5.asPoint())

# Selection
indices = [1, 0, 2, 1, 2]
print("selection=", sample5.select(indices))
indices = list(range(len(sample5)))
print("selection(id)=", sample5.select(indices))

# html conversion
print(sample1._repr_html_())

# check conversion
v1 = [ot.Point([12.0]), ot.Point([1.0])]
v2 = [ot.Point([13.0]), ot.Point([2.0])]
try:
    s = ot.Sample([v1, v2])
except Exception:
    print("ok")

try:
    # uneven points
    s = ot.Sample([[1.0, 2.0], [5.0]])
except Exception:
    print("ok")

# getMarginal by identifiers
sample = ot.Normal(4).getSample(5)
assert sample.getMarginal(["X3", "X1"]).getDescription() == [
    "X3",
    "X1",
], "wrong marginal"

# argsort
sample = ot.Sample(
    [[-1.0, 1.0, 0.0], [-1.0, 1.0, 1.0], [-1.0, 0.0, 1.0], [-1.0, 0.0, -1.0]]
)
indices = sample.argsort()
print("indices = ", indices)
assert indices == [3, 2, 0, 1]
indices = sample.argsort(False)
print("indices = ", indices)
assert indices == [1, 0, 2, 3]


# this should not crash


def centeredSumOfProduct(sampleA, sampleB, center):
    sampleA -= center
    sampleB -= center
    return sampleA.asPoint().dot(sampleB.asPoint())


YA = ot.Sample(10, 1)
center = 0.0
sumOfSquaresOfG = centeredSumOfProduct(YA, YA, center)

# html conversion of a Sample with fractional numbers
sampleFractional = ot.Sample(
    [
        [0.1234567890, 0.3333333333],
        [-123456.78901, 0.6666666666],
        [1.2345678901e10, -1.2345678901e-12],
    ]
)
print(sampleFractional._repr_html_())



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Instantiate one distribution object
first = -1.5
ll = [1.0, 0.7, 1.2, 0.9]
h = [0.5, 1.5, 3.5, 2.5]
t = [first]
f = list()
for i in range(len(ll)):
    t.append(t[i] + ll[i])
    f.append(h[i] * ll[i])
distribution = ot.Histogram(t, f)
print("Distribution ", repr(distribution))
print("Distribution ", distribution)
distribution = ot.Histogram(first, ll, h)
print("Distribution ", repr(distribution))
print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", repr(oneRealization))

# Define a point
point = ot.Point(distribution.getDimension(), 1.0)
print("Point= ", repr(point))

# derivative of PDF with regards its arguments
DDF = distribution.computeDDF(point)
print("ddf     =", repr(DDF))

# PDF value
LPDF = distribution.computeLogPDF(point)
print("log pdf=%.6f" % LPDF)
PDF = distribution.computePDF(point)
print("pdf     =%.6f" % PDF)

# derivative of the PDF with regards the parameters of the distribution
CDF = distribution.computeCDF(point)
print("cdf=%.6f" % CDF)
CCDF = distribution.computeComplementaryCDF(point)
print("ccdf=%.6f" % CCDF)

# quantile
quantile = distribution.computeQuantile(0.95)
print("quantile=", repr(quantile))
print("cdf(quantile)=%.6f" % distribution.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f"
    % distribution.computeSurvivalFunction(inverseSurvival)
)
print("entropy=%.6f" % distribution.computeEntropy())

# Confidence regions
interval, threshold = distribution.computeMinimumVolumeIntervalWithMarginalProbability(
    0.95
)
print("Minimum volume interval=", interval)
print("threshold=", ot.Point(1, threshold))
levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
print("Minimum volume level set=", levelSet)
print("beta=", ot.Point(1, beta))
interval, beta = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(
    0.95
)
print("Bilateral confidence interval=", interval)
print("beta=", ot.Point(1, beta))
(
    interval,
    beta,
) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(0.95, False)
print("Unilateral confidence interval (lower tail)=", interval)
print("beta=", ot.Point(1, beta))
(
    interval,
    beta,
) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(0.95, True)
print("Unilateral confidence interval (upper tail)=", interval)
print("beta=", ot.Point(1, beta))

mean = distribution.getMean()
print("mean=", repr(mean))
covariance = distribution.getCovariance()
print("covariance=", repr(covariance))
parameters = distribution.getParametersCollection()
print("parameters=", repr(parameters))
print("Standard representative=", distribution.getStandardRepresentative())

ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.skipParameters()
validation.skipMinimumVolumeLevelSet()
validation.run()



import openturns as ot

assert ot.Interval() != ot.ExpectationSimulationAlgorithm()
assert not ot.Interval() == ot.ExpectationSimulationAlgorithm()

assert ot.Interval() != ot.Sample([[1]])
assert not ot.Interval() == ot.Sample([[1]])

assert ot.Laplace() != ot.UsualRandomVector(ot.Normal())
assert not ot.Laplace() == ot.UsualRandomVector(ot.Normal())



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()

# Set hmat
ot.ResourceMap.SetAsString("KrigingAlgorithm-LinearAlgebra", "HMAT")


def test_one_input_one_output():
    sampleSize = 6
    dimension = 1

    f = ot.SymbolicFunction(["x0"], ["x0 * sin(x0)"])

    X = ot.Sample(sampleSize, dimension)
    X2 = ot.Sample(sampleSize, dimension)
    for i in range(sampleSize):
        X[i, 0] = 3.0 + i
        X2[i, 0] = 2.5 + i
    X[0, 0] = 1.0
    X[1, 0] = 3.0
    X2[0, 0] = 2.0
    X2[1, 0] = 4.0
    Y = f(X)
    f(X2)

    # create algorithm
    basis = ot.ConstantBasisFactory(dimension).build()
    covarianceModel = ot.SquaredExponential([1e-02], [4.50736])

    algo = ot.KrigingAlgorithm(X, Y, covarianceModel, basis)
    algo.run()

    # perform an evaluation
    result = algo.getResult()

    ott.assert_almost_equal(result.getMetaModel()(X), Y, 1e-2)

    # Kriging variance is 0 on learning points
    covariance = result.getConditionalCovariance(X)
    ot.Point(covariance.getImplementation())
    ot.Point(sampleSize * sampleSize)
    ott.assert_almost_equal(covariance, ot.Matrix(sampleSize, sampleSize), 0.0, 1e-1)

    # Covariance per marginal & extract variance component
    coll = result.getConditionalMarginalCovariance(X)
    var = [mat[0, 0] for mat in coll]
    ott.assert_almost_equal(var, [0] * sampleSize, 0.0, 1e-1)

    # Variance per marginal
    var = result.getConditionalMarginalVariance(X)
    ott.assert_almost_equal(var, ot.Sample(sampleSize, 1), 0.0, 1e-1)


# Test 2
def test_two_inputs_one_output():
    # Kriging use case
    inputDimension = 2

    # Learning data
    levels = [8, 5]
    box = ot.Box(levels)
    inputSample = box.generate()
    # Scale each direction
    inputSample *= 10.0

    model = ot.SymbolicFunction(["x", "y"], ["cos(0.5*x) + sin(y)"])
    outputSample = model(inputSample)

    # Validation
    sampleSize = 10
    inputValidSample = ot.JointDistribution(2 * [ot.Uniform(0, 10.0)]).getSample(
        sampleSize
    )
    outputValidSample = model(inputValidSample)

    # 2) Definition of exponential model
    # The parameters have been calibrated using TNC optimization
    # and AbsoluteExponential models
    covarianceModel = ot.SquaredExponential([5.33532, 2.61534], [1.61536])

    # 3) Basis definition
    basis = ot.ConstantBasisFactory(inputDimension).build()
    # Kriging algorithm
    algo = ot.KrigingAlgorithm(inputSample, outputSample, covarianceModel, basis)
    algo.run()
    result = algo.getResult()
    # Get meta model
    metaModel = result.getMetaModel()
    metaModel(inputValidSample)

    # 4) Errors
    # Interpolation
    ott.assert_almost_equal(outputSample, metaModel(inputSample), 1, 3.0e-5)

    # 5) Kriging variance is 0 on learning points
    covariance = result.getConditionalCovariance(inputSample)
    ott.assert_almost_equal(covariance, ot.SquareMatrix(len(inputSample)), 0.0, 1e-3)

    # Covariance per marginal & extract variance component
    coll = result.getConditionalMarginalCovariance(inputSample)
    var = [mat[0, 0] for mat in coll]
    ott.assert_almost_equal(var, [0] * len(var), 0.0, 1e-3)

    # Variance per marginal
    var = result.getConditionalMarginalVariance(inputSample)
    ott.assert_almost_equal(var, ot.Sample(inputSample.getSize(), 1), 0.0, 1e-3)
    # Estimation
    ott.assert_almost_equal(
        outputValidSample, metaModel(inputValidSample), 1.0e-1, 1e-1
    )


def test_stationary_fun():
    # fix https://github.com/openturns/openturns/issues/1861
    ot.RandomGenerator.SetSeed(0)
    rho = ot.SymbolicFunction("tau", "exp(-abs(tau))*cos(2*pi_*abs(tau))")
    model = ot.StationaryFunctionalCovarianceModel([1], [1], rho)
    x = ot.Normal().getSample(20)
    y = x + ot.Normal(0, 0.1).getSample(20)

    algo = ot.KrigingAlgorithm(x, y, model, ot.LinearBasisFactory().build())
    algo.run()
    result = algo.getResult()
    variance = result.getConditionalMarginalVariance(x)
    ott.assert_almost_equal(variance, ot.Sample(len(x), 1), 2e-6, 2e-6)


if __name__ == "__main__":
    test_one_input_one_output()
    test_two_inputs_one_output()
    test_stationary_fun()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


def quadM(m, n):
    res = ot.Matrix(m, n)
    for i in range(m):
        for j in range(n):
            res[i, j] = (i + 1.0) ** (j + 1.0)
    return res


def testQR(m, n, full, keep):
    matrix1 = quadM(m, n)
    print("M=", matrix1)
    if keep:
        Q, R = matrix1.computeQR(full)
    else:
        Q, R = matrix1.computeQRInPlace(full)
    print("full=", full, "keep=", keep)
    print("Q= ", Q)
    print("R=", R)
    print("Q*R=", Q * R)
    if keep:
        print("M2=", matrix1)


# Square case
matrix1 = quadM(3, 3)
matrix1.setName("matrix1")
print("matrix1 = ", repr(matrix1))

result1 = matrix1.computeSingularValues()
print("svd (svd only)= ", repr(result1))

result1, u, v = matrix1.computeSVD(True)
print("svd (svd + U, V full)= ", repr(result1))
result1, u, v = matrix1.computeSVD(False)
print("svd (svd + U, V small)= ", repr(result1), ", U=", repr(u), ", v=", repr(v))

# Rectangular case, m < n
matrix1 = quadM(3, 5)
matrix1.setName("matrix1")
print("matrix1 = ", repr(matrix1))

result1 = matrix1.computeSingularValues()
print("svd (svd only)= ", repr(result1))

result1, u, v = matrix1.computeSVD(True)
print("svd (svd + U, V full)= ", repr(result1))
result1, u, v = matrix1.computeSVD(False)
print("svd (svd + U, V small)= ", repr(result1), ", U=", repr(u), ", v=", repr(v))

# Rectangular case, m > n
matrix1 = quadM(5, 3)
matrix1.setName("matrix1")
print("matrix1 = ", repr(matrix1))

result1 = matrix1.computeSingularValues()
print("svd (svd only)= ", repr(result1))

result1, u, v = matrix1.computeSVD(True)
print("svd (svd + U, V full)= ", repr(result1))
# result1, u, v = matrix1.computeSVD(False)
# print "svd (svd + U, V small)= ", repr(result1), ", U=", repr(u), ",
# v=", repr(v)

for iFull in range(2):
    for iKeep in range(2):
        testQR(3, 3, iFull == 1, iKeep == 1)
        testQR(3, 5, iFull == 1, iKeep == 1)
        testQR(5, 3, iFull == 1, iKeep == 1)



"""
Fit a non parametric copula
===========================
"""

# %%
# In this example we are going to estimate a Normal copula from a sample using non parametric representations.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create data
R = ot.CorrelationMatrix(2)
R[1, 0] = 0.4
copula = ot.NormalCopula(R)
sample = copula.getSample(30)

# %%
# Estimate a Normal copula using :class:`~openturns.BernsteinCopulaFactory`
distribution = ot.BernsteinCopulaFactory().build(sample)

# %%
# Draw fitted distribution
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# Estimate a Normal copula using :class:`~openturns.KernelSmoothing`
distribution = ot.KernelSmoothing().build(sample).getCopula()
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import math

ot.TESTPREAMBLE()

# First, a smooth function
f = ot.SymbolicFunction("x", "sin(x)")
a = -2.5
b = 4.5
# Default parameters
algo = ot.GaussKronrod()
rules = ot.GaussKronrod.GetRules()
for i in range(len(rules)):
    algo.setRule(rules[i])
    print("Algo=", algo)
    # High-level interface
    error = ot.Point()
    value = algo.integrate(f, ot.Interval(a, b), error)[0]
    ref = math.cos(a) - math.cos(b)
    print(
        "value=%.6f" % value,
        ", ref=%.6f" % ref,
        ", true error below bound? ",
        abs(ref - value) < algo.getMaximumError(),
        ", estimated error below bound? ",
        error[0] < algo.getMaximumError(),
    )
    # Low-level interface
    # ai = Point()
    # bi = Point()
    # fi = Sample()
    # ei = Point()
    # value2 = algo.integrate(f, a, b, error, ai, bi, fi, ei)[0]
    # ai.add(b)
    # g = f.draw(a, b, 512)
    # lower = Cloud(ai, Point(ai.getDimension()))
    # lower.setColor("magenta")
    # g.add(lower)

# Second, a piecewise smooth function
f = ot.SymbolicFunction("x", "abs(sin(x))")
a = -2.5
b = 4.5
algo = ot.GaussKronrod()
rules = ot.GaussKronrod.GetRules()
rules[0] = ot.GaussKronrod.GetRuleFromName("G3K7")

for i in range(len(rules)):
    algo.setRule(rules[i])
    print("Algo=", algo)
    error = ot.Point()
    value = algo.integrate(f, ot.Interval(a, b), error)[0]
    ref = 4.0 + math.cos(b) - math.cos(a)
    print(
        "value=%.6f" % value,
        ", ref=%.6f" % ref,
        ", true error below bound? ",
        abs(ref - value) < algo.getMaximumError(),
        ", estimated error below bound? ",
        error[0] < algo.getMaximumError(),
    )
    # Low-level interface
    # ai = Point()
    # bi = Point()
    # fi = Sample()
    # ei = Point()
    # value2 = algo.integrate(f, a, b, error, ai, bi, fi, ei)[0]
    # print "ai=", ai
    # print "bi=", bi
    # print "fi=", fi
    # print "ei=", ei
    # ai.add(b)
    # g = f.draw(a, b, 512)
    # lower = Cloud(ai, Point(ai.getDimension()))
    # lower.setColor("magenta")
    # g.add(lower)



"""
Estimate a probability with Monte Carlo
=======================================
"""

# %%
# In this example we estimate a probability by means of a simulation algorithm, the Monte-Carlo algorithm.
# To do this, we need the classes `MonteCarloExperiment` and `ProbabilitySimulationAlgorithm`.
# We consider the :ref:`axial stressed beam <use-case-stressed-beam>` example.


# %%
from openturns.usecases import stressed_beam
import openturns as ot


# %%
# We load the model from the usecases module :
sm = stressed_beam.AxialStressedBeam()

# %%
# We get the joint distribution of the parameters.
distribution = sm.distribution

# %%
# The model is also stored in the data class :
model = sm.model

# %%
# We create the event whose probability we want to estimate.

# %%
vect = ot.RandomVector(distribution)
G = ot.CompositeRandomVector(model, vect)
event = ot.ThresholdEvent(G, ot.Less(), 0.0)

# %%
# Create a Monte Carlo algorithm.

# %%
experiment = ot.MonteCarloExperiment()
algo = ot.ProbabilitySimulationAlgorithm(event, experiment)
algo.setMaximumCoefficientOfVariation(0.05)
algo.setMaximumOuterSampling(int(1e5))
algo.setKeepSample(True)
algo.run()

# %%
# Retrieve results.

# %%
result = algo.getResult()
probability = result.getProbabilityEstimate()
print("Pf=", probability)

# %%
# Print the size of used sample and compare it with the budget size (1e5)
print(f"Sample used to get the probability = {len(algo.getOutputSample())}")
